# sage_setup: distribution = sagemath-graphs
r"""
KnotInfo database

This module contains the class :class:`KnotInfoDataBase` and auxiliary classes
for it, which serves as an interface to the lists of named knots and links provided
at the web-pages `KnotInfo <https://knotinfo.math.indiana.edu/>`__ and
`LinkInfo <https://linkinfo.sitehost.iu.edu>`__.

To use the database, you need to install the optional :ref:`database_knotinfo
<spkg_database_knotinfo>` package by the Sage command ::

    sage -i database_knotinfo

EXAMPLES::

    sage: # optional - database_knotinfo
    sage: from sage.databases.knotinfo_db import KnotInfoDataBase
    sage: ki_db = KnotInfoDataBase()
    sage: ki_db
    <sage.databases.knotinfo_db.KnotInfoDataBase object at ...>

AUTHORS:

- Sebastian Oehms (2020-08): initial version
"""
##############################################################################
#       Copyright (C) 2020 Sebastian Oehms <seb.oehms@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
##############################################################################
import os
from enum import Enum

from sage.structure.sage_object import SageObject
from sage.structure.unique_representation import UniqueRepresentation
from sage.misc.persist import save, load
from sage.misc.verbose import verbose
from sage.misc.cachefunc import cached_method

columns_white_list = ['knot_atlas_anon', 'knotilus_page_anon']
columns_black_list = ['homfly_polynomial_old']


class KnotInfoColumnTypes(Enum):
    r"""
    Enum class to specify if a column from the table of knots and links provided
    at the web-pages `KnotInfo <https://knotinfo.math.indiana.edu/>`__ and
    `LinkInfo <https://linkinfo.sitehost.iu.edu>`__.  is used for knots only,
    links only or both.

    EXAMPLES::

        sage: from sage.databases.knotinfo_db import KnotInfoColumnTypes
        sage: [col_type for col_type in KnotInfoColumnTypes]
        [<KnotInfoColumnTypes.OnlyKnots: 'K'>,
        <KnotInfoColumnTypes.OnlyLinks: 'L'>,
        <KnotInfoColumnTypes.KnotsAndLinks: 'B'>]
    """
    OnlyKnots = 'K'       # column that is only used in the KnotInfo table
    OnlyLinks = 'L'       # column that is only used in the LinkInfo table
    KnotsAndLinks = 'B'       # column that is only used in both tables


class KnotInfoColumns(Enum):
    r"""
    Enum class to select a column from the table of knots and links provided
    at the web-pages `KnotInfo <https://knotinfo.math.indiana.edu/>`__ and
    `LinkInfo <https://linkinfo.sitehost.iu.edu>`__.

    EXAMPLES::

        sage: from sage.databases.knotinfo_db import KnotInfoDataBase
        sage: ki_db = KnotInfoDataBase()
        sage: cols = ki_db.columns(); cols
        <enum 'Columns'>
        sage: from sage.databases.knotinfo_db import KnotInfoColumns
        sage: isinstance(cols.name, KnotInfoColumns)
        True

        sage: def only_links(c):
        ....:     return c.column_type() == c.types.OnlyLinks
        sage: [c.column_name() for c in cols if only_links(c)]  # optional - database_knotinfo
        ['Name - Unoriented',
         'Orientation',
         'Unoriented Rank',
         'PD Notation (vector)',
         'PD Notation (KnotTheory)',
         'Braid Notation',
         'Quasipositive Braid',
         'Multivariable Alexander Polynomial',
         'HOMFLYPT Polynomial',
         'Khovanov Polynomial',
         'Unoriented',
         'Arc Notation',
         'Linking Matrix',
         'Rolfsen Name',
         'Components',
         'DT code',
         'Splitting Number',
         'Nullity',
         'Unlinking Number',
         'Weak Splitting Number']
    """
    @property
    def types(self):
        r"""
        Return :class:`KnotInfoColumnTypes` to be used for checks.

        EXAMPLES::

            sage: from sage.databases.knotinfo_db import KnotInfoDataBase
            sage: ki_db = KnotInfoDataBase()
            sage: cols = ki_db.columns()
            sage: cols.dt_code.column_type() == cols.dt_code.types.OnlyLinks
            True
        """
        return KnotInfoColumnTypes

    def column_name(self):
        r"""
        Return the name of ``self`` displayed on the KnotInfo web-page.

        EXAMPLES::

            sage: from sage.databases.knotinfo_db import KnotInfoDataBase
            sage: ki_db = KnotInfoDataBase()
            sage: cols = ki_db.columns()
            sage: cols.dt_code.column_name()
            'DT code'
        """
        return self.value[0]

    def column_type(self):
        r"""
        Return the type of ``self``. That is an instance of ``Enum``
        :class:`KnotInfoColumnTypes`.

        EXAMPLES::

            sage: from sage.databases.knotinfo_db import KnotInfoDataBase
            sage: ki_db = KnotInfoDataBase()
            sage: cols = ki_db.columns()
            sage: cols.homfly_polynomial.column_type()
            <KnotInfoColumnTypes.OnlyKnots: 'K'>
            sage: cols.homflypt_polynomial.column_type()
            <KnotInfoColumnTypes.OnlyLinks: 'L'>
            sage: cols.name.column_type()
            <KnotInfoColumnTypes.KnotsAndLinks: 'B'>
        """
        return self.value[1]

    def description_webpage(self, new=0, autoraise=True):
        r"""
        Launch the description page of ``self`` in the standard web browser.

        EXAMPLES::

            sage: from sage.databases.knotinfo_db import KnotInfoDataBase
            sage: ki_db = KnotInfoDataBase()
            sage: cols = ki_db.columns()
            sage: cols.pd_notation.description_webpage()            # not tested
            True
            sage: cols.homflypt_polynomial.description_webpage()    # not tested
            True
        """
        import webbrowser
        if self.column_type() == self.types.OnlyLinks:
            url = KnotInfoFilename.links.description_url(self)
        else:
            url = KnotInfoFilename.knots.description_url(self)
        return webbrowser.open(url, new=new, autoraise=autoraise)


class KnotInfoFilename(Enum):
    r"""
    Enum for the different data files. The following choices are possible:

    - ``knots`` -- contains the data from KnotInfo
    - ``links`` -- contains the data for proper links from LinkInfo

    Examples::

        sage: from sage.databases.knotinfo_db import KnotInfoDataBase
        sage: ki_db = KnotInfoDataBase()
        sage: ki_db.filename
        <enum 'KnotInfoFilename'>
    """

    def url(self):
        r"""
        Return the URL to download the data from the web-page.

        Examples::

            sage: from sage.databases.knotinfo_db import KnotInfoDataBase
            sage: ki_db = KnotInfoDataBase()
            sage: ki_db.filename.knots.url()
            'https://knotinfo.math.indiana.edu/'
        """
        if self == KnotInfoFilename.knots:
            return self.value[0]
        else:
            return self.value[0]

    def excel(self):
        r"""
        Return the Excel-file name to download the data from the web-page.

        Examples::

            sage: from sage.databases.knotinfo_db import KnotInfoDataBase
            sage: ki_db = KnotInfoDataBase()
            sage: ki_db.filename.knots.excel()
            'knotinfo_data_complete.xls'
        """
        if self == KnotInfoFilename.knots:
            return '%s.xls' % (self.value[1])
        else:
            return '%s.xlsx' % (self.value[1])

    def csv(self):
        r"""
        Return the file name under which the data from the web-page
        are stored as csv file.

        Examples::

            sage: from sage.databases.knotinfo_db import KnotInfoDataBase
            sage: ki_db = KnotInfoDataBase()
            sage: ki_db.filename.knots.csv()
            'knotinfo_data_complete.csv'
        """
        return '%s.csv' % (self.value[1])

    def num_knots(self, version):
        r"""
        Return the file name under which the number of knots is stored
        in an sobj-file.

        Examples::

            sage: from sage.databases.knotinfo_db import KnotInfoDataBase
            sage: ki_db = KnotInfoDataBase()
            sage: ki_db.filename.knots.num_knots('21.7')
            'num_knots_21.7.sobj'
        """
        return 'num_knots_%s.sobj' % version

    def sobj_row(self):
        r"""
        Return the file name under which the row-data of the csv-File
        is stored as python dictionary in a sobj-file.

        Examples::

            sage: from sage.databases.knotinfo_db import KnotInfoDataBase
            sage: ki_db = KnotInfoDataBase()
            sage: ki_db.filename.knots.sobj_row()
            'row_dict.sobj'
        """
        return 'row_dict.sobj'

    def sobj_column(self):
        r"""
        Return the file name under which the column-data of the csv-File
        is stored as python dictionary in a sobj-file.

        Examples::

            sage: from sage.databases.knotinfo_db import KnotInfoDataBase
            sage: ki_db = KnotInfoDataBase()
            sage: ki_db.filename.knots.sobj_column()
            'column_dict.sobj'
        """
        return 'column_dict.sobj'

    def sobj_data(self, column):
        r"""
        Return the file name under which the data of the given
        column is stored as python list in a sobj-file.

        Examples::

            sage: from sage.databases.knotinfo_db import KnotInfoDataBase
            sage: ki_db = KnotInfoDataBase()
            sage: ki_db.filename.knots.sobj_data(ki_db.columns().braid_notation)
            'knotinfo_braid_notation'
        """
        if column.column_type() == column.types.OnlyLinks:
            return 'linkinfo_%s' % (column.name)
        else:
            return 'knotinfo_%s' % (column.name)

    def description_url(self, column):
        r"""
        Return the url of the description page of the given column.

        Examples::

            sage: from sage.databases.knotinfo_db import KnotInfoDataBase
            sage: ki_db = KnotInfoDataBase()
            sage: ki_db.filename.knots.description_url(ki_db.columns().braid_notation)
            'https://knotinfo.math.indiana.edu/descriptions/braid_notation.html'
        """
        return '%sdescriptions/%s.html' % (self.url(), column.name)

    def diagram_url(self, fname, single=False):
        r"""
        Return the url of the diagram page of the given link.

        Examples::

            sage: from sage.databases.knotinfo_db import KnotInfoDataBase
            sage: ki_db = KnotInfoDataBase()
            sage: ki_db.filename.knots.diagram_url('3_1-50.png')
            'https://knotinfo.math.indiana.edu/diagram_display.php?3_1-50.png'
            sage: ki_db.filename.knots.diagram_url('3_1', single=True)
            'https://knotinfo.math.indiana.edu/diagrams/3_1'
        """
        if single:
            return '%sdiagrams/%s' % (self.url(), fname)
        else:
            return '%sdiagram_display.php?%s' % (self.url(), fname)

    knots = ['https://knotinfo.math.indiana.edu/', 'knotinfo_data_complete']
    links = ['https://linkinfo.sitehost.iu.edu/',  'linkinfo_data_complete']


#----------------------------------------------------------------------------------------------------------------------------
# Class to provide data for knots and links from the KnotInfo web-page
#----------------------------------------------------------------------------------------------------------------------------
class KnotInfoDataBase(SageObject, UniqueRepresentation):
    r"""
    Database interface to KnotInfo.

    The original data are obtained from KnotInfo web-page (URL see the example
    below). In order to have these data installed during the build process as
    a sage-package they are converted as csv files into a tarball. This tarball
    has been created using the method :meth:`create_spkg_tarball`.

    EXAMPLES::

        sage: from sage.databases.knotinfo_db import KnotInfoDataBase
        sage: ki_db = KnotInfoDataBase()
        sage: ki_db.filename.knots
        <KnotInfoFilename.knots: ['https://knotinfo.math.indiana.edu/',
                                  'knotinfo_data_complete']>
    """

    filename = KnotInfoFilename

    def __init__(self, install=False):
        r"""
        Python constructor.

        EXAMPLES::

            sage: from sage.databases.knotinfo_db import KnotInfoDataBase
            sage: ki_db = KnotInfoDataBase()
            sage: ki_db.filename.links
            <KnotInfoFilename.links: ['https://linkinfo.sitehost.iu.edu/',
                                      'linkinfo_data_complete']>
        """
        # some constants
        self._names_column = 'name'
        self._components_column = 'components'
        self._knot_prefix = 'K'

        self._knot_list = None
        self._link_list = None
        self._demo = None
        self._num_knots = None

        from sage.features.databases import DatabaseKnotInfo
        from sage.env import DOT_SAGE
        self._feature = DatabaseKnotInfo()
        self._sobj_path = os.path.join(DOT_SAGE, 'knotinfo')

    def create_filecache(self, force=False):
        r"""
        Create the internal files containing the database.

        INPUT:

        - ``force`` -- boolean (default: ``False``); if set to ``True`` the
          existing file-cache is overwritten

        EXAMPLES::

            sage: from sage.databases.knotinfo_db import KnotInfoDataBase
            sage: ki_db = KnotInfoDataBase()
            sage: ki_db.create_filecache()    # optional - database_knotinfo
        """
        if not self._feature.is_present():
            return

        if os.path.isdir(self._sobj_path):
            # if it exists then remove it if it belongs to an older version of
            # the database or should be reset by the user because it is damaged.
            test_version = os.path.join(self._sobj_path, self.filename.knots.num_knots(self.version()))
            if force or not os.path.isfile(test_version):
                import shutil
                shutil.rmtree(self._sobj_path)

        from sage.misc.temporary_file import atomic_dir
        with atomic_dir(self._sobj_path) as d:
            sobj_path = d.name
            num_knots_file = os.path.join(sobj_path, self.filename.knots.num_knots(self.version()))
            knot_list = self.knot_list()
            num_knots = len(knot_list) - 1
            save(num_knots, num_knots_file)
            self._num_knots = num_knots
            self._create_col_dict_sobj(sobj_path=sobj_path)
            self._create_data_sobj(sobj_path=sobj_path)
        return

    def version(self):
        r"""
        Return the version of the database currently installed on the device.

        .. NOTE::

            The development of the original databases on the KnotInfo and
            LinkInfo web-pages is in a continuous flow. The installed version
            can be behind the current available state of these databases. Every
            month a cronjob on the
            `GitHub repository <https://github.com/soehms/database_knotinfo/>`__
            searches for differences and creates a new release on
            `PyPI <https://pypi.org/project/database-knotinfo/>`__ in case of
            success.

            If you note that your version is behind the version on PyPI
            and would like to have Sage working with that release you should
            first try to upgrade using ``sage -i database_knotinfo``. If this
            is not successful even though you are on the latest Sage release
            please create an issue for that in the GitHub repository.

        EXAMPLES::

            sage: from sage.databases.knotinfo_db import KnotInfoDataBase
            sage: ki_db = KnotInfoDataBase()
            sage: ki_db.version()   >= '2021.10.1'   # optional database_knotinfo
            True
        """
        self._feature.require()
        from database_knotinfo import version
        return version()

    def demo_version(self):
        r"""
        Return whether the KnotInfo databases are installed completely or
        just the demo version is used.

        EXAMPLES::

            sage: from sage.databases.knotinfo_db import KnotInfoDataBase
            sage: ki_db = KnotInfoDataBase()
            sage: ki_db.demo_version()       # optional - database_knotinfo
            False
        """
        if self._demo is None:
            if self._feature.is_present():
                num_knots_file = os.path.join(self._sobj_path, self.filename.knots.num_knots(self.version()))
                from builtins import FileNotFoundError
                try:
                    self._num_knots = load(num_knots_file)
                except FileNotFoundError:
                    self.create_filecache()
                self._demo = False
            else:
                self._demo = True
                self._num_knots = len([v for v in row_demo_sample.values() if v[1] == 1])
        return self._demo

    def knot_list(self):
        r"""
        Return the KnotInfo table rows as Python list.

        EXAMPLES::

            sage: from sage.databases.knotinfo_db import KnotInfoDataBase
            sage: ki_db = KnotInfoDataBase()
            sage: len(ki_db.knot_list())  # not tested (just used on installation)
        """
        if self._knot_list:
            return self._knot_list

        from database_knotinfo import link_list
        self._knot_list = link_list()
        return self._knot_list

    def link_list(self):
        r"""
        Return the LinkInfo table rows as Python list.

        EXAMPLES::

            sage: from sage.databases.knotinfo_db import KnotInfoDataBase
            sage: ki_db = KnotInfoDataBase()
            sage: len(ki_db.link_list())  # not tested (just used on installation)
        """
        if self._link_list:
            return self._link_list

        from database_knotinfo import link_list
        self._link_list = link_list(proper_links=True)
        return self._link_list

    def _create_col_dict_sobj(self, sobj_path=None):
        r"""
        Create ``sobj`` files containing the number of knots and a dictionary
        for the columns of the table.

        EXAMPLES::

            sage: from sage.databases.knotinfo_db import KnotInfoDataBase
            sage: ki_db = KnotInfoDataBase()
            sage: ki_db._create_col_dict_sobj() # not tested (used on installation)
        """
        knot_list = self.knot_list()
        knot_column_names = knot_list[0]

        link_list = self.link_list()
        link_column_names = link_list[0]

        column_dict = {}

        if not sobj_path:
            sobj_path = self._sobj_path

        # ----------------------------------------------------------------
        # Columns that exist for knots and links
        # ----------------------------------------------------------------
        for col in knot_column_names:

            name = knot_column_names[col]
            if not name and col not in columns_white_list:
                # not of interest
                continue

            if col in columns_black_list:
                # not of interest
                continue

            col_type = KnotInfoColumnTypes.OnlyKnots
            if col in link_column_names:
                col_type = KnotInfoColumnTypes.KnotsAndLinks
            column_dict[col] = [name, col_type]

        # ----------------------------------------------------------------
        # Columns that exist for links only
        # ----------------------------------------------------------------
        for col in link_column_names:

            name = link_column_names[col]
            if not name and col not in columns_white_list:
                # not of interest
                continue

            if col in columns_black_list:
                # not of interest
                continue

            if col in knot_column_names:
                # already used
                continue

            col_type = KnotInfoColumnTypes.OnlyLinks
            column_dict[col] = [name, col_type]

        save(column_dict, '%s/%s' % (sobj_path, self.filename.knots.sobj_column()))

    def _create_data_sobj(self, sobj_path=None):
        r"""
        Create ``sobj`` files containing the contents of the whole table.
        To each column there is created one file containing a list of
        strings giving the entries of the database table.

        The length of these lists depends on the type of the corresponding
        column. If a column is used in both tables
        (``KnotInfoColumnTypes.KnotsAndLinks``) the list of proper links
        is appended to the list of knots.  In both other cases the lenght
        of the list corresponds to the number of listed knots and proper
        links respectively.

        EXAMPLES::

            sage: from sage.databases.knotinfo_db import KnotInfoDataBase
            sage: ki_db = KnotInfoDataBase()
            sage: ki_db._create_data_sobj() # not tested (just used on installation)
        """
        knot_list = self.knot_list()
        link_list = self.link_list()
        len_knots = len(knot_list)
        len_links = len(link_list)

        row_dict = {}

        if not sobj_path:
            sobj_path = self._sobj_path

        # ----------------------------------------------------------------
        # Columns that exist for knots and links
        # ----------------------------------------------------------------
        column_dict = load('%s/%s' % (sobj_path, self.filename.knots.sobj_column()))
        cols = KnotInfoColumns('ColsTemp', column_dict)
        for col in cols:
            val_list = []

            if col.column_type() != col.types.OnlyLinks:
                for i in range(1, len_knots):
                    if col.name == self._names_column:
                        row_dict[self._knot_prefix + knot_list[i][col.name]] = [i - 1, 1]

                    val_list.append(knot_list[i][col.name])

            if col.column_type() != col.types.OnlyKnots:
                for i in range(1, len_links):
                    if col.name == self._names_column:
                        link_name = link_list[i][col.name]
                        link_name = link_name.replace('{', '_')
                        link_name = link_name.replace(',', '_')
                        link_name = link_name.replace('}', '')

                        num_comp = int(link_list[i][self._components_column])
                        row_dict[link_name] = [i + len_knots - 2, num_comp]

                    val_list.append(link_list[i][col.name])

            if val_list:
                save(val_list, '%s/%s' % (sobj_path, self.filename.knots.sobj_data(col)))

        save(row_dict,    '%s/%s' % (sobj_path, self.filename.knots.sobj_row()))

    @cached_method
    def columns(self):
        r"""
        Return the columns ot the database table as instances of enum class
        :class:`KnotInfoColumns`.

        EXAMPLES::

            sage: from sage.databases.knotinfo_db import KnotInfoDataBase
            sage: ki_db = KnotInfoDataBase()
            sage: cols = ki_db.columns()
            sage: [col.column_name() for col in cols if col.name.startswith('pd')]   # optional - database_knotinfo
            ['PD Notation', 'PD Notation (vector)', 'PD Notation (KnotTheory)']
        """
        column_dict = self.read_column_dict()
        return KnotInfoColumns('Columns', column_dict)

    # -------------------------------------------------------------------------------------------------------------
    # read the dictionary for the column names from sobj-file
    # -------------------------------------------------------------------------------------------------------------
    @cached_method
    def read_column_dict(self):
        r"""
        Read the dictionary for the column names from the according sobj-file.

        OUTPUT: a Python dictionary containing the column names and types

        EXAMPLES::

            sage: from sage.databases.knotinfo_db import KnotInfoDataBase
            sage: ki_db = KnotInfoDataBase()
            sage: len(ki_db.read_column_dict()) > 120       # optional - database_knotinfo
            True
        """
        if self.demo_version():
            return column_demo_sample
        sobj_path = self._sobj_path
        filename = self.filename.knots.sobj_column()
        return load('%s/%s' % (sobj_path, filename))

    # -------------------------------------------------------------------------------------------------------------
    # read the dictionary for the row names that is the knot and link names from sobj-file
    # -------------------------------------------------------------------------------------------------------------
    @cached_method
    def read_row_dict(self):
        r"""
        Read the dictionary for the row names that is the knot and link names
        from the according sobj-file

        OUTPUT:

        A python dictionary containing the names of the knots and links
        together with their table index and the corresponding number of
        components

        EXAMPLES::

            sage: from sage.databases.knotinfo_db import KnotInfoDataBase
            sage: ki_db = KnotInfoDataBase()
            sage: ki_db.read_row_dict()['K7_1']
            [8, 1]
        """
        if self.demo_version():
            return row_demo_sample
        sobj_path = self._sobj_path
        filename = self.filename.knots.sobj_row()
        return load('%s/%s' % (sobj_path, filename))

    # -------------------------------------------------------------------------------------------------------------
    # return a dictionary to obtain the original name to a row_dict key
    # -------------------------------------------------------------------------------------------------------------
    @cached_method
    def row_names(self):
        r"""
        Return a dictionary to obtain the original name to a row_dict key.

        OUTPUT: a Python dictionary containing the names of the knots and links
        together with their original names from the database

        EXAMPLES::

            sage: from sage.databases.knotinfo_db import KnotInfoDataBase
            sage: ki_db = KnotInfoDataBase()
            sage: ki_db.row_names()['K7_1']      # optional - database_knotinfo
            '7_1'
        """
        row_dict = self.read_row_dict()
        names = self.read(self.columns().name)
        return {k:names[v[0]] for k, v in row_dict.items()}

    # -------------------------------------------------------------------------------------------------------------
    # read the number of knots contained in the database (without proper links) from the according sobj-file.
    # -------------------------------------------------------------------------------------------------------------
    def read_num_knots(self):
        r"""
        Read the number of knots contained in the database (without
        proper links) from the according sobj-file.

        OUTPUT: integer

        EXAMPLES::

            sage: from sage.databases.knotinfo_db import KnotInfoDataBase
            sage: ki_db = KnotInfoDataBase()
            sage: ki_db.read_num_knots()              # optional - database_knotinfo
            12966
        """
        if not self._num_knots:
            self.demo_version()
        return self._num_knots

    # -------------------------------------------------------------------------------------------------------------
    # read an sobj-file obtained from KnotInfo database
    # -------------------------------------------------------------------------------------------------------------
    @cached_method
    def read(self, column):
        r"""
        Access a column of KnotInfo / LinkInfo.

        INPUT:

        - ``column`` -- instance of enum :class:`KnotInfoColumns`
          to select the data to be read in

        OUTPUT: a Python list containing the data corresponding to the column

        EXAMPLES::

            sage: from sage.databases.knotinfo_db import KnotInfoDataBase
            sage: ki_db = KnotInfoDataBase()
        """
        if not isinstance(column, KnotInfoColumns):
            raise TypeError('column must be an instance of enum %s' % (KnotInfoColumns))

        if self.demo_version():
            return data_demo_sample[column]

        sobj_path = self._sobj_path
        if column.column_type() == column.types.OnlyLinks:
            filename = self.filename.links.sobj_data(column)
        else:
            filename = self.filename.knots.sobj_data(column)

        verbose('loading data library %s ...' % (filename))
        res = load('%s/%s' % (sobj_path, filename))
        verbose('... finished!')

        return res

    def _test_database(self, **options):
        r"""
        Method used by TestSuite. Performs :meth:`KnotInfoBase.is_recoverable`.

        EXAMPLES::

            sage: from sage.databases.knotinfo_db import KnotInfoDataBase
            sage: ki_db = KnotInfoDataBase()
            sage: TestSuite(ki_db).run()    # optional - database_knotinfo, long time, indirect doctest
        """
        from sage.knots.knotinfo import KnotInfo
        from sage.misc.misc import some_tuples
        tester = options['tester']
        max_samples = tester._max_samples
        if not max_samples:
            max_samples = 20
        l = list(KnotInfo)
        sample = some_tuples(l, 1, len(l), max_samples=max_samples)
        tester.assertTrue(all(L.is_recoverable(unique=False) for L, in sample))


column_demo_sample = {
    'name':                 ['Name',                 KnotInfoColumnTypes.KnotsAndLinks],
    'name_unoriented':      ['Name - Unoriented',    KnotInfoColumnTypes.OnlyLinks],
    'dt_notation':          ['DT Notation',          KnotInfoColumnTypes.OnlyKnots],
    'gauss_notation':       ['Gauss Notation',       KnotInfoColumnTypes.KnotsAndLinks],
    'pd_notation':          ['PD Notation',          KnotInfoColumnTypes.OnlyKnots],
    'pd_notation_vector':   ['PD Notation (vector)', KnotInfoColumnTypes.OnlyLinks],
    'crossing_number':      ['Crossing Number',      KnotInfoColumnTypes.KnotsAndLinks],
    'braid_index':          ['Braid Index',          KnotInfoColumnTypes.OnlyKnots],
    'braid_length':         ['Braid Length',         KnotInfoColumnTypes.OnlyKnots],
    'braid_notation':       ['Braid Notation',       KnotInfoColumnTypes.KnotsAndLinks],
    'braid_notation_old':   ['Braid Notation',       KnotInfoColumnTypes.OnlyLinks],
    'alternating':          ['Alternating',          KnotInfoColumnTypes.KnotsAndLinks],
    'alexander_polynomial': ['Alexander',            KnotInfoColumnTypes.OnlyKnots],
    'jones_polynomial':     ['Jones',                KnotInfoColumnTypes.KnotsAndLinks],
    'conway_polynomial':    ['Conway',               KnotInfoColumnTypes.KnotsAndLinks],
    'homfly_polynomial':    ['HOMFLY',               KnotInfoColumnTypes.OnlyKnots],
    'homflypt_polynomial':  ['HOMFLYPT Polynomial',  KnotInfoColumnTypes.OnlyLinks],
    'kauffman_polynomial':  ['Kauffman',             KnotInfoColumnTypes.KnotsAndLinks],
    'khovanov_polynomial':  ['Khovanov',             KnotInfoColumnTypes.OnlyLinks],
    'khovanov_unreduced_integral_polynomial': ['KH Unred Z Poly', KnotInfoColumnTypes.OnlyKnots],
    'khovanov_reduced_integral_polynomial': ['KH Red Z Poly', KnotInfoColumnTypes.OnlyKnots],
    'khovanov_reduced_rational_polynomial': ['KH Red Q Poly', KnotInfoColumnTypes.OnlyKnots],
    'khovanov_reduced_mod2_polynomial': ['KH Red Mod2 Poly', KnotInfoColumnTypes.OnlyKnots],
    'khovanov_odd_integral_polynomial': ['KH Odd Red Z Poly', KnotInfoColumnTypes.OnlyKnots],
    'khovanov_odd_rational_polynomial': ['KH Odd Red Q Poly', KnotInfoColumnTypes.OnlyKnots],
    'khovanov_odd_mod2_polynomial': ['KH Red Odd Mod2 Poly', KnotInfoColumnTypes.OnlyKnots],
    'determinant':          ['Determinant',          KnotInfoColumnTypes.KnotsAndLinks],
    'positive':             ['Positive',             KnotInfoColumnTypes.OnlyKnots],
    'fibered':              ['Fibered',              KnotInfoColumnTypes.OnlyKnots],
    'unoriented':           ['Unoriented',           KnotInfoColumnTypes.OnlyLinks],
    'symmetry_type':        ['Symmetry Type',        KnotInfoColumnTypes.OnlyKnots],
    'geometric_type':       ['Geometric Type',       KnotInfoColumnTypes.OnlyKnots],
    'cosmetic_crossing':    ['Cosmetic Crossing',    KnotInfoColumnTypes.OnlyKnots],
    'width':                ['Width',                KnotInfoColumnTypes.OnlyKnots],
    'arc_notation':         ['Arc Notation',         KnotInfoColumnTypes.OnlyLinks],
    'dt_code':              ['DT code',              KnotInfoColumnTypes.OnlyLinks]
}


row_demo_sample = {
    'K0_1':   [0, 1],
    'K3_1':   [1, 1],
    'K4_1':   [2, 1],
    'K5_1':   [3, 1],
    'K5_2':   [4, 1],
    'K6_1':   [5, 1],
    'K6_2':   [6, 1],
    'K6_3':   [7, 1],
    'K7_1':   [8, 1],
    'K7_2':   [9, 1],
    'L2a1_0': [10, 2],
    'L2a1_1': [11, 2],
    'L4a1_0': [12, 2],
    'L4a1_1': [13, 2],
    'L5a1_0': [14, 2],
    'L5a1_1': [15, 2],
    'L6a1_0': [16, 2],
    'L6a1_1': [17, 2],
    'L6a2_0': [18, 2],
    'L6a2_1': [19, 2]
}

db = KnotInfoDataBase()
dc = db.columns()


data_demo_sample = {
    dc.name: ['0_1', '3_1', '4_1', '5_1', '5_2', '6_1', '6_2', '6_3', '7_1', '7_2',
              'L2a1{0}', 'L2a1{1}', 'L4a1{0}', 'L4a1{1}', 'L5a1{0}', 'L5a1{1}',
              'L6a1{0}', 'L6a1{1}', 'L6a2{0}', 'L6a2{1}', 'L6a3{0}'
             ],
    dc.name_unoriented: ['L2a1', 'L2a1', 'L4a1', 'L4a1', 'L5a1', 'L5a1', 'L6a1', 'L6a1', 'L6a2', 'L6a2', 'L6a3'],
    dc.crossing_number: ['0', '3', '4', '5', '5', '6', '6', '6', '7', '7', '2', '2', '4', '4', '5', '5', '6', '6', '6', '6', '6'],
    dc.braid_notation: [
        '',
        '{1,1,1}',
        '{1,-2,1,-2}',
        '{1,1,1,1,1}',
        '{1,1,1,2,-1,2}',
        '{1,1,2,-1,-3,2,-3}',
        '{1,1,1,-2,1,-2}',
        '{1,1,-2,1,-2,-2}',
        '{1,1,1,1,1,1,1}',
        '{1,1,1,2,-1,2,3,-2,3}',
        '{2, {-1, -1}}',
        '{2, {1, 1}}',
        '{3, {-2, -2, -1, 2, -1}}',
        '{2, {1, 1, 1, 1}}',
        '{3, {-1, 2, -1, 2, -1}}',
        '{3, {-1, 2, -1, 2, -1}}',
        '{4, {1, -2, 3, -2, 1, -2, -3, -2}}',
        '{3, {2, 2, 2, 1, 1, -2, 1}}',
        '{3, {-1, 2, -1, -2, -2, -1, -1}}',
        '{3, {1, -2, 1, 2, 2, 1, 1}}',
        '{2, {-1, -1, -1, -1, -1, -1}}'
        ],
    dc.braid_notation_old: [
        '{2, {-1, -1}}',
        '{2, {1, 1}}',
        '{4, {1, -2, 3, -2, -1, -2, -3, -2}}',
        '{2, {1, 1, 1, 1}}',
        '{3, {-1, 2, -1, 2, -1}}',
        '{3, {-1, 2, -1, 2, -1}}',
        '{4, {1, -2, 3, -2, 1, -2, -3, -2}}',
        '{4, {1, 2, 3, 2, 2, -1, 2, 2, -3, 2}}',
        '{4, {1, -2, -2, -2, 3, -2, -1, -2, -3, -2}}',
        '{4, {1, 2, -3, 2, -1, 2, 3, 2, 2, 2}}',
        '{2, {-1, -1, -1, -1, -1, -1}}'
        ],
    dc.braid_index: ['1', '2', '3', '2', '3', '4', '3', '3', '2', '4'],
    dc.braid_length: ['', '3', '4', '5', '6', '7', '6', '6', '7', '9'],
    dc.determinant: ['0', '3', '5', '5', '7', '9', '11', '13', '7', '11', '2', '2', '4', '4', '8', '8', '12', '12', '10', '10', '6'],
    dc.positive: ['', 'Y', 'N', 'Y', 'Y', 'N', 'N', 'N', 'Y', 'Y'],
    dc.fibered: ['', 'Y', 'Y', 'Y', 'N', 'N', 'Y', 'Y', 'Y', 'N'],
    dc.unoriented: ['Y', 'N', 'Y', 'N', 'Y', 'N', 'Y', 'N', 'Y', 'N', 'Y'],
    dc.pd_notation: [
        '',
        '[[1,5,2,4],[3,1,4,6],[5,3,6,2]]',
        '[[4,2,5,1],[8,6,1,5],[6,3,7,4],[2,7,3,8]]',
        '[[2,8,3,7],[4,10,5,9],[6,2,7,1],[8,4,9,3],[10,6,1,5]]',
        '[[1,5,2,4],[3,9,4,8],[5,1,6,10],[7,3,8,2],[9,7,10,6]]',
        '[[1,7,2,6],[3,10,4,11],[5,3,6,2],[7,1,8,12],[9,4,10,5],[11,9,12,8]]',
        '[[1,8,2,9],[3,11,4,10],[5,1,6,12],[7,2,8,3],[9,7,10,6],[11,5,12,4]]',
        '[[4,2,5,1],[8,4,9,3],[12,9,1,10],[10,5,11,6],[6,11,7,12],[2,8,3,7]]',
        '[[1,9,2,8],[3,11,4,10],[5,13,6,12],[7,1,8,14],[9,3,10,2],[11,5,12,4],[13,7,14,6]]',
        '[[2,10,3,9],[4,14,5,13],[6,12,7,11],[8,2,9,1],[10,8,11,7],[12,6,13,5],[14,4,1,3]]',
        ],
    dc.pd_notation_vector: [
        '{{4, 1, 3, 2}, {2, 3, 1, 4}}',
        '{{4, 2, 3, 1}, {2, 4, 1, 3}}',
        '{{6, 1, 7, 2}, {8, 3, 5, 4}, {2, 5, 3, 6}, {4, 7, 1, 8}}',
        '{{6, 2, 7, 1}, {8, 4, 5, 3}, {2, 8, 3, 7}, {4, 6, 1, 5}}',
        '{{6, 1, 7, 2}, {10, 7, 5, 8}, {4, 5, 1, 6}, {2, 10, 3, 9}, {8, 4, 9, 3}}',
        '{{8, 2, 9, 1}, {10, 7, 5, 8}, {4, 10, 1, 9}, {2, 5, 3, 6}, {6, 3, 7, 4}}',
        '{{6, 1, 7, 2}, {10, 3, 11, 4}, {12, 8, 5, 7}, {8, 12, 9, 11}, {2, 5, 3, 6}, {4, 9, 1, 10}}',
        '{{10, 2, 11, 1}, {6, 4, 7, 3}, {12, 10, 5, 9}, {8, 6, 9, 5}, {2, 12, 3, 11}, {4, 8, 1, 7}}',
        '{{8, 1, 9, 2}, {12, 5, 7, 6}, {10, 3, 11, 4}, {4, 11, 5, 12}, {2, 7, 3, 8}, {6, 9, 1, 10}}',
        '{{10, 2, 11, 1}, {12, 6, 7, 5}, {8, 4, 9, 3}, {4, 8, 5, 7}, {2, 12, 3, 11}, {6, 10, 1, 9}}',
        '{{8, 1, 9, 2}, {2, 9, 3, 10}, {10, 3, 11, 4}, {12, 5, 7, 6}, {6, 7, 1, 8}, {4, 11, 5, 12}}'
        ],
    dc.dt_notation: [
        '',
        '[4, 6, 2]',
        '[4, 6, 8, 2]',
        '[6, 8, 10, 2, 4]',
        '[4, 8, 10, 2, 6]',
        '[4, 8, 12, 10, 2, 6]',
        '[4, 8, 10, 12, 2, 6]',
        '[4, 8, 10, 2, 12, 6]',
        '[8, 10, 12, 14, 2, 4, 6]',
        '[4, 10, 14, 12, 2, 8, 6]'
        ],
    dc.dt_code: [
        '[{4}, {2}]',
        '[{4}, {2}]',
        '[{6, 8}, {2, 4}]',
        '[{6, 8}, {4, 2}]',
        '[{6, 8}, {4, 10, 2}]',
        '[{8, 6}, {2, 10, 4}]',
        '[{6, 10}, {2, 12, 4, 8}]',
        '[{10, 6}, {8, 4, 12, 2}]',
        '[{8, 10, 12}, {2, 6, 4}]',
        '[{10, 8, 12}, {4, 6, 2}]',
        '[{8, 10, 12}, {6, 2, 4}]'
        ],
    dc.gauss_notation: [
        '',
        '{1, -2, 3, -1, 2, -3}',
        '{-1, 2, -3, 1, -4, 3, -2, 4}',
        '{-1, 2, -3, 4, -5, 1, -2, 3, -4, 5}',
        '{1, -2, 3, -1, 4, -5, 2, -3, 5, -4}',
        '{1, -2, 3, -4, 2, -1, 5, -6, 4, -3, 6, -5}',
        '{1, -2, 3, -4, 5, -6, 2, -1, 6, -3, 4, -5}',
        '{-1, 2, -3, 1, -4, 5, -2, 3, -6, 4, -5, 6}',
        '{1, -2, 3, -4, 5, -6, 7, -1, 2, -3, 4, -5, 6, -7}',
        '{-1, 2, -3, 4, -5, 6, -7, 1, -2, 7, -6, 5, -4, 3}',
        '{{1, -2}, {2, -1}}',
        '{{1, -2}, {2, -1}}',
        '{{1, -3, 2, -4}, {3, -1, 4, -2}}',
        '{{1, -3, 2, -4}, {4, -1, 3, -2}}',
        '{{1, -4, 5, -3}, {3, -1, 2, -5, 4, -2}}',
        '{{1, -4, 5, -3}, {4, -5, 2, -1, 3, -2}}',
        '{{1, -5, 2, -6}, {5, -1, 3, -4, 6, -2, 4, -3}}',
        '{{1, -5, 2, -6}, {4, -2, 6, -4, 3, -1, 5, -3}}',
        '{{1, -5, 3, -4, 2, -6}, {5, -1, 6, -3, 4, -2}}',
        '{{1, -5, 3, -4, 2, -6}, {4, -3, 6, -1, 5, -2}}',
        '{{1, -2, 3, -6, 4, -5}, {5, -1, 2, -3, 6, -4}}'
        ],
    dc.arc_notation: [
        '{{4, 2}, {3, 1}, {4, 2}, {1, 3}}',
        '{{2, 4}, {3, 1}, {2, 4}, {3, 1}}',
        '{{6, 4}, {3, 5}, {4, 2}, {1, 3}, {2, 6}, {5, 1}}',
        '{{3, 6}, {2, 5}, {6, 4}, {1, 3}, {5, 2}, {4, 1}}',
        '{{6, 2}, {1, 4}, {3, 5}, {4, 7}, {2, 6}, {7, 3}, {5, 1}}',
        '{{3, 5}, {6, 4}, {5, 2}, {7, 3}, {1, 6}, {2, 7}, {4, 1}}',
        '{{8, 4}, {3, 5}, {4, 2}, {6, 3}, {5, 7}, {1, 6}, {2, 8}, {7, 1}}',
        '{{2, 8}, {1, 7}, {8, 4}, {5, 3}, {4, 2}, {3, 6}, {7, 5}, {6, 1}}',
        '{{8, 3}, {2, 7}, {3, 1}, {4, 8}, {5, 2}, {6, 4}, {7, 5}, {1, 6}}',
        '{{3, 8}, {2, 7}, {8, 4}, {1, 3}, {5, 2}, {4, 6}, {7, 5}, {6, 1}}',
        '{{8, 2}, {1, 3}, {2, 4}, {3, 5}, {4, 6}, {5, 7}, {6, 8}, {7, 1}}'
        ],
    dc.alternating: ['Y', 'Y', 'Y', 'Y', 'Y', 'Y', 'Y', 'Y', 'Y', 'Y', 'Y', 'Y', 'Y', 'Y', 'Y', 'Y', 'Y', 'Y', 'Y', 'Y', 'Y'],
    dc.symmetry_type: [
        '',
        'reversible',
        'fully amphicheiral',
        'reversible',
        'reversible',
        'reversible',
        'reversible',
        'fully amphicheiral',
        'reversible',
        'reversible'
        ],
    dc.geometric_type: [
        '',
        'torus knot T(2,3)',
        'hyperbolic',
        'torus knot T(2,5)',
        'hyperbolic',
        'hyperbolic',
        'hyperbolic',
        'hyperbolic',
        'torus knot T(2,7)',
        'hyperbolic'],
    dc.cosmetic_crossing: ['', 'N', 'N', 'N', 'N', 'N', 'N', 'N', 'N', 'N'],
    dc.homfly_polynomial: [
        '',
        '(2*v^2-v^4)+v^2*z^2',
        '(v^(-2)-1+v^2)-z^2',
        '(3*v^4-2*v^6)+(4*v^4-v^6)*z^2+v^4*z^4',
        '(v^2+v^4-v^6)+(v^2+v^4)*z^2',
        '(v^(-2)-v^2+v^4)+(-1-v^2)*z^2',
        '(2-2*v^2+v^4)+(1-3*v^2+v^4)*z^2-v^2*z^4',
        '(-v^(-2)+3-v^2)+(-v^(-2)+3-v^2)*z^2+z^4',
        '(4*v^6-3*v^8)+(10*v^6-4*v^8)*z^2+(6*v^6-v^8)*z^4+v^6*z^6',
        '(v^2+v^6-v^8)+(v^2+v^4+v^6)*z^2'
        ],
    dc.homflypt_polynomial: [
        '1/(v^3*z)-1/(v*z)-z/v',
        'v/z-v^3/z + v*z',
        '1/(v^5*z)-1/(v^3*z)-z/v^3-z/v',
        'v^3/z-v^5/z + 3*v^3*z-v^5*z + v^3*z^3',
        '1/(v*z)-v/z-z/v^3 + (2*z)/v-v*z + z^3/v',
        '1/(v*z)-v/z-z/v^3 + (2*z)/v-v*z + z^3/v',
        '1/(v^5*z)-1/(v^3*z)-(2*z)/v^3 + z/v-v*z + z^3/v',
        'v^3/z-v^5/z + 2*v^3*z + v^5*z-v^7*z + v^3*z^3 + v^5*z^3',
        '1/(v^7*z)-1/(v^5*z) + z/v^7-(2*z)/v^5-(2*z)/v^3-z^3/v^5-z^3/v^3',
        'v^5/z-v^7/z + 2*v^3*z + 2*v^5*z-v^7*z + v^3*z^3 + v^5*z^3',
        '1/(v^7*z)-1/(v^5*z) + (3*z)/v^7-(6*z)/v^5 + z^3/v^7-(5*z^3)/v^5-z^5/v^5'
        ],
    dc.kauffman_polynomial: [
        '',
        '(-a^(-4)-2*a^(-2))*z^(0)+ (a^(-5)+ a^(-3))*z^(1)+ (a^(-4)+ a^(-2))*z^(2)',
        '(-a^(-2)-1-a^2)*z^(0)+ (-a^(-1)-a)*z^(1)+ (a^(-2)+ 2+ a^2)*z^(2)+ (a^(-1)+ a)*z^(3)',
        '(2*a^(-6)+ 3*a^(-4))*z^(0)+ (a^(-9)-a^(-7)-2*a^(-5))*z^(1)+ (a^(-8)-3*a^(-6)-4*a^(-4))*z^(2)+ (a^(-7)+ a^(-5))*z^(3)+ (a^(-6)+ a^(-4))*z^(4)',
        '(a^(-6)+ a^(-4)-a^(-2))*z^(0)+ (-2*a^(-7)-2*a^(-5))*z^(1)+ (-2*a^(-6)-a^(-4)+ a^(-2))*z^(2)+ (a^(-7)+ 2*a^(-5)+ a^(-3))*z^(3)+ (a^(-6)+ a^(-4))*z^(4)',
        '(a^(-4)+ a^(-2)-a^2)*z^(0)+ (2*a^(-3)+ 2*a^(-1))*z^(1)+ (-3*a^(-4)-4*a^(-2)+ a^2)*z^(2)+ (-3*a^(-3)-2*a^(-1)+ a)*z^(3)+ (a^(-4)+ 2*a^(-2)+ 1)*z^(4)+ (a^(-3)+ a^(-1))*z^(5)',
        '(a^(-4)+ 2*a^(-2)+ 2)*z^(0)+ (-a^(-5)-a^(-3))*z^(1)+ (a^(-6)-2*a^(-4)-6*a^(-2)-3)*z^(2)+ (2*a^(-5)-2*a^(-1))*z^(3)+ (2*a^(-4)+ 3*a^(-2)+ 1)*z^(4)+ (a^(-3)+ a^(-1))*z^(5)',
        '(a^(-2)+ 3+ a^2)*z^(0)+ (-a^(-3)-2*a^(-1)-2*a-a^3)*z^(1)+ (-3*a^(-2)-6-3*a^2)*z^(2)+ (a^(-3)+ a^(-1)+ a+ a^3)*z^(3)+ (2*a^(-2)+ 4+ 2*a^2)*z^(4)+ (a^(-1)+ a)*z^(5)',
        '(-3*a^(-8)-4*a^(-6))*z^(0)+ (a^(-13)-a^(-11)+ a^(-9)+ 3*a^(-7))*z^(1)+ (a^(-12)-2*a^(-10)+ 7*a^(-8)+ 10*a^(-6))*z^(2)+ (a^(-11)-3*a^(-9)-4*a^(-7))*z^(3)+ (a^(-10)-5*a^(-8)-6*a^(-6))*z^(4)+ (a^(-9)+ a^(-7))*z^(5)+ (a^(-8)+ a^(-6))*z^(6)',
        '(-a^(-8)-a^(-6)-a^(-2))*z^(0)+ (3*a^(-9)+ 3*a^(-7))*z^(1)+ (4*a^(-8)+ 3*a^(-6)+ a^(-2))*z^(2)+ (-4*a^(-9)-6*a^(-7)-a^(-5)+ a^(-3))*z^(3)+ (-4*a^(-8)-3*a^(-6)+ a^(-4))*z^(4)+ (a^(-9)+ 2*a^(-7)+ a^(-5))*z^(5)+ (a^(-8)+ a^(-6))*z^(6)',
        'a^2-a/z-a^3/z + a*z + a^3*z',
        'a^(-2)-1/(a^3*z)-1/(a*z) + z/a^3 + z/a',
        '-a^4 + a^3/z + a^5/z + a*z-2*a^3*z-3*a^5*z + a^2*z^2 + a^4*z^2 + a^3*z^3 + a^5*z^3',
        '-a^(-4) + 1/(a^5*z) + 1/(a^3*z) + z/a^7-(2*z)/a^5-(3*z)/a^3 + z^2/a^6 + z^2/a^4 + z^3/a^5 + z^3/a^3',
        '-1 + 1/(a*z) + a/z-(2*z)/a-4*a*z-2*a^3*z-z^2 + a^4*z^2 + z^3/a + 3*a*z^3 + 2*a^3*z^3 + z^4 + a^2*z^4',
        '-1 + 1/(a*z) + a/z-(2*z)/a-4*a*z-2*a^3*z-z^2 + a^4*z^2 + z^3/a + 3*a*z^3 + 2*a^3*z^3 + z^4 + a^2*z^4',
        '-a^4 + a^3/z + a^5/z-z/a-a^3*z-2*a^5*z-3*z^2-3*a^2*z^2 + z^3/a + a^5*z^3 + 2*z^4 + 3*a^2*z^4 + a^4*z^4 + a*z^5 + a^3*z^5',
        '-a^(-4) + 1/(a^5*z) + 1/(a^3*z)-z/a^9-z/a^5-(2*z)/a^3-(3*z^2)/a^8-(3*z^2)/a^6 + z^3/a^9 + z^3/a^3 + (2*z^4)/a^8 + (3*z^4)/a^6 + z^4/a^4 + z^5/a^7 + z^5/a^5',
        'a^6-a^5/z-a^7/z-2*a^3*z + 3*a^5*z + 3*a^7*z-2*a^9*z-a^4*z^2-2*a^6*z^2-a^8*z^2 + a^3*z^3-2*a^5*z^3-2*a^7*z^3 + a^9*z^3 + a^4*z^4 + 2*a^6*z^4 + a^8*z^4 + a^5*z^5 + a^7*z^5',
        'a^(-6)-1/(a^7*z)-1/(a^5*z)-(2*z)/a^9 + (3*z)/a^7 + (3*z)/a^5-(2*z)/a^3-z^2/a^8-(2*z^2)/a^6-z^2/a^4 + z^3/a^9-(2*z^3)/a^7-(2*z^3)/a^5 + z^3/a^3 + z^4/a^8 + (2*z^4)/a^6 + z^4/a^4 + z^5/a^7 + z^5/a^5',
        'a^6-a^5/z-a^7/z + 6*a^5*z + 4*a^7*z-a^9*z + a^11*z-3*a^6*z^2-2*a^8*z^2 + a^10*z^2-5*a^5*z^3-4*a^7*z^3 + a^9*z^3 + a^6*z^4 + a^8*z^4 + a^5*z^5 + a^7*z^5'
        ],
    dc.jones_polynomial: [
        '1',
        't+ t^3-t^4',
        't^(-2)-t^(-1)+ 1-t+ t^2',
        't^2+ t^4-t^5+ t^6-t^7',
        't-t^2+ 2*t^3-t^4+ t^5-t^6',
        't^(-2)-t^(-1)+ 2-2*t+ t^2-t^3+ t^4',
        't^(-1)-1+ 2*t-2*t^2+ 2*t^3-2*t^4+ t^5',
        '-t^(-3)+ 2*t^(-2)-2*t^(-1)+ 3-2*t+ 2*t^2-t^3',
        't^3+ t^5-t^6+ t^7-t^8+ t^9-t^10',
        't-t^2+ 2*t^3-2*t^4+ 2*t^5-t^6+ t^7-t^8',
        '-x^(-5)-x^(-1)',
        '-x-x^5',
        '-x^(-9)-x^(-5) + x^(-3)-x^(-1)',
        '-x^3-x^7 + x^9-x^11',
        'x^(-7)-2/x^5 + x^(-3)-2/x + x-x^3',
        'x^(-7)-2/x^5 + x^(-3)-2/x + x-x^3',
        '-x^(-9) + x^(-7)-3/x^5 + 2/x^3-2/x + 2*x-x^3',
        '-x^3 + x^5-3*x^7 + 2*x^9-2*x^11 + 2*x^13-x^15',
        '-x^(-15) + x^(-13)-2/x^11 + 2/x^9-2/x^7 + x^(-5)-x^(-3)',
        '-x^3 + x^5-2*x^7 + 2*x^9-2*x^11 + x^13-x^15',
        '-x^(-17) + x^(-15)-x^(-13) + x^(-11)-x^(-9)-x^(-5)'
        ],
    dc.alexander_polynomial: [
        '1',
        '1-t+ t^2',
        '1-3*t+ t^2',
        '1-t+ t^2-t^3+ t^4',
        '2-3*t+ 2*t^2',
        '2-5*t+ 2*t^2',
        '1-3*t+ 3*t^2-3*t^3+ t^4',
        '1-3*t+ 5*t^2-3*t^3+ t^4',
        '1-t+ t^2-t^3+ t^4-t^5+ t^6',
        '3-5*t+ 3*t^2'],
    dc.conway_polynomial: [
        '1',
        '1+z^2',
        '1-z^2',
        '1+3*z^2+z^4',
        '1+2*z^2',
        '1-2*z^2',
        '1-z^2-z^4',
        '1+z^2+z^4',
        '1+6*z^2+5*z^4+z^6',
        '1+3*z^2',
        '-z',
        'z',
        '-2*z',
        '2*z + z^3',
        'z^3',
        'z^3',
        '-2*z + z^3',
        '2*z + 2*z^3',
        '-3*z-2*z^3',
        '3*z + 2*z^3',
        '-3*z-4*z^3-z^5'],
    dc.khovanov_polynomial: [
        '1 + q^(-2) + 1/(q^6*t^2) + 1/(q^4*t^2)',
        '1 + q^2 + q^4*t^2 + q^6*t^2',
        '1 + q^(-2) + 1/(q^10*t^4) + 1/(q^8*t^4) + 1/(q^6*t^2) + 1/(q^2*t)',
        'q^2 + q^4 + q^6*t^2 + q^10*t^3 + q^10*t^4 + q^12*t^4',
        '2 + 2/q^2 + 1/(q^8*t^3) + 1/(q^6*t^2) + 1/(q^4*t^2) + 1/(q^2*t) + t + q^4*t^2',
        '2 + 2/q^2 + 1/(q^8*t^3) + 1/(q^6*t^2) + 1/(q^4*t^2) + 1/(q^2*t) + t + q^4*t^2',
        '1 + 2/q^2 + 1/(q^10*t^4) + 1/(q^8*t^4) + 1/(q^8*t^3) + 2/(q^6*t^2) + 1/(q^4*t^2) + 2/(q^2*t) + t + q^2*t + q^4*t^2',
        'q^2 + q^4 + q^4*t + 2*q^6*t^2 + q^8*t^2 + 2*q^10*t^3 + 2*q^10*t^4 + q^12*t^4 + q^12*t^5 + q^14*t^5 + q^16*t^6',
        'q^(-4) + q^(-2) + 1/(q^16*t^6) + 1/(q^14*t^6) + 1/(q^14*t^5) + 1/(q^12*t^4) + 1/(q^10*t^4) + 1/(q^10*t^3) + 1/(q^8*t^3) + 1/(q^8*t^2) + 1/(q^6*t^2) + 1/(q^4*t)',
        'q^2 + q^4 + q^4*t + q^6*t^2 + q^8*t^2 + q^8*t^3 + q^10*t^3 + q^10*t^4 + q^12*t^4 + q^14*t^5 + q^14*t^6 + q^16*t^6',
        'q^(-6) + q^(-4) + 1/(q^18*t^6) + 1/(q^16*t^6) + 1/(q^16*t^5) + 1/(q^12*t^4) + 1/(q^12*t^3) + 1/(q^8*t^2)'],
    dc.khovanov_unreduced_integral_polynomial: [
        '',
        'q + q^(3) + t^(2) q^(5) + t^(3) q^(9) + t^(3) q^(7) T^(2)',
        't^(-2) q^(-5) + t^(-1) q^(-1) + q^(-1) + q + t q + t^(2) q^(5) + t^(-1) q^(-3) T^(2) + t^(2) q^(3) T^(2)',
        'q^(3) + q^(5) + t^(2) q^(7) + t^(3) q^(11) + t^(4) q^(11) + t^(5) q^(15) + t^(3) q^(9) T^(2) + t^(5) q^(13) T^(2)',
        'q + q^(3) + t q^(3) + t^(2) q^(5) + t^(2) q^(7) + t^(3) q^(9) + t^(4) q^(9) + t^(5) q^(13) + t^(2) q^(5) T^(2) + t^(3) q^(7) T^(2) + t^(5) q^(11) T^(2)',
        't^(-2) q^(-5) + t^(-1) q^(-1) + 2 q^(-1) + q + t q + t q^(3) + t^(2) q^(5) + t^(3) q^(5) + t^(4) q^(9) + t^(-1) q^(-3) T^(2) + t q T^(2) + t^(2) q^(3) T^(2) + t^(4) q^(7) T^(2)',
        't^(-2) q^(-3) + t^(-1) q + 2 q + q^(3) + t q^(3) + t q^(5) + t^(2) q^(5) + t^(2) q^(7) + t^(3) q^(7) + t^(3) q^(9) + t^(4) q^(11) + t^(-1) q^(-1) T^(2) + t q^(3) T^(2) + t^(2) q^(5) T^(2) + t^(3) q^(7) T^(2) + t^(4) q^(9) T^(2)',
        't^(-3) q^(-7) + t^(-2) q^(-5) + t^(-2) q^(-3) + t^(-1) q^(-3) + t^(-1) q^(-1) + 2 q^(-1) + 2 q + t q + t q^(3) + t^(2) q^(3) + t^(2) q^(5) + t^(3) q^(7) + t^(-2) q^(-5) T^(2) + t^(-1) q^(-3) T^(2) + q^(-1) T^(2) + t q T^(2) + t^(2) q^(3) T^(2) + t^(3) q^(5) T^(2)',
        'q^(5) + q^(7) + t^(2) q^(9) + t^(3) q^(13) + t^(4) q^(13) + t^(5) q^(17) + t^(6) q^(17) + t^(7) q^(21) + t^(3) q^(11) T^(2) + t^(5) q^(15) T^(2) + t^(7) q^(19) T^(2)',
        'q + q^(3) + t q^(3) + t^(2) q^(5) + t^(2) q^(7) + t^(3) q^(7) + t^(3) q^(9) + t^(4) q^(9) + t^(4) q^(11) + t^(5) q^(13) + t^(6) q^(13) + t^(7) q^(17) + t^(2) q^(5) T^(2) + t^(3) q^(7) T^(2) + t^(4) q^(9) T^(2) + t^(5) q^(11) T^(2) + t^(7) q^(15) T^(2)'],
    dc.khovanov_reduced_integral_polynomial: [
        '',
        'q^(2) + t^(2) q^(6) + t^(3) q^(8)',
        't^(-2) q^(-4) + t^(-1) q^(-2) + 1 + t q^(2) + t^(2) q^(4)',
        'q^(4) + t^(2) q^(8) + t^(3) q^(10) + t^(4) q^(12) + t^(5) q^(14)',
        'q^(2) + t q^(4) + 2 t^(2) q^(6) + t^(3) q^(8) + t^(4) q^(10) + t^(5) q^(12)',
        't^(-2) q^(-4) + t^(-1) q^(-2) + 2 + 2 t q^(2) + t^(2) q^(4) + t^(3) q^(6) + t^(4) q^(8)',
        't^(-2) q^(-2) + t^(-1) + 2 q^(2) + 2 t q^(4) + 2 t^(2) q^(6) + 2 t^(3) q^(8) + t^(4) q^(10)',
        't^(-3) q^(-6) + 2 t^(-2) q^(-4) + 2 t^(-1) q^(-2) + 3 + 2 t q^(2) + 2 t^(2) q^(4) + t^(3) q^(6)',
        'q^(6) + t^(2) q^(10) + t^(3) q^(12) + t^(4) q^(14) + t^(5) q^(16) + t^(6) q^(18) + t^(7) q^(20)',
        'q^(2) + t q^(4) + 2 t^(2) q^(6) + 2 t^(3) q^(8) + 2 t^(4) q^(10) + t^(5) q^(12) + t^(6) q^(14) + t^(7) q^(16)'],
    dc.khovanov_reduced_rational_polynomial: [
        '',
        ' q^(2) + t^(2) q^(6) + t^(3) q^(8)',
        't^(-2) q^(-4) + t^(-1) q^(-2) + 1 + t q^(2) + t^(2) q^(4)',
        ' q^(4) + t^(2) q^(8) + t^(3) q^(10) + t^(4) q^(12) + t^(5) q^(14)',
        ' q^(2) + t q^(4) + 2 t^(2) q^(6) + t^(3) q^(8) + t^(4) q^(10) + t^(5) q^(12)',
        't^(-2) q^(-4) + t^(-1) q^(-2) + 2 + 2 t q^(2) + t^(2) q^(4) + t^(3) q^(6) + t^(4) q^(8)',
        't^(-2) q^(-2) + t^(-1) + 2 q^(2) + 2 t q^(4) + 2 t^(2) q^(6) + 2 t^(3) q^(8) + t^(4) q^(10)',
        't^(-3) q^(-6) + 2 t^(-2) q^(-4) + 2 t^(-1) q^(-2) + 3 + 2 t q^(2) + 2 t^(2) q^(4) + t^(3) q^(6)',
        ' q^(6) + t^(2) q^(10) + t^(3) q^(12) + t^(4) q^(14) + t^(5) q^(16) + t^(6) q^(18) + t^(7) q^(20)',
        ' q^(2) + t q^(4) + 2 t^(2) q^(6) + 2 t^(3) q^(8) + 2 t^(4) q^(10) + t^(5) q^(12) + t^(6) q^(14) + t^(7) q^(16)'],
    dc.khovanov_reduced_mod2_polynomial: [
        '',
        ' q^(2) + t^(2) q^(6) + t^(3) q^(8)',
        't^(-2) q^(-4) + t^(-1) q^(-2) + 1 + t q^(2) + t^(2) q^(4)',
        ' q^(4) + t^(2) q^(8) + t^(3) q^(10) + t^(4) q^(12) + t^(5) q^(14)',
        ' q^(2) + t q^(4) + 2 t^(2) q^(6) + t^(3) q^(8) + t^(4) q^(10) + t^(5) q^(12)',
        't^(-2) q^(-4) + t^(-1) q^(-2) + 2 + 2 t q^(2) + t^(2) q^(4) + t^(3) q^(6) + t^(4) q^(8)',
        't^(-2) q^(-2) + t^(-1) + 2 q^(2) + 2 t q^(4) + 2 t^(2) q^(6) + 2 t^(3) q^(8) + t^(4) q^(10)',
        't^(-3) q^(-6) + 2 t^(-2) q^(-4) + 2 t^(-1) q^(-2) + 3 + 2 t q^(2) + 2 t^(2) q^(4) + t^(3) q^(6)',
        ' q^(6) + t^(2) q^(10) + t^(3) q^(12) + t^(4) q^(14) + t^(5) q^(16) + t^(6) q^(18) + t^(7) q^(20)',
        ' q^(2) + t q^(4) + 2 t^(2) q^(6) + 2 t^(3) q^(8) + 2 t^(4) q^(10) + t^(5) q^(12) + t^(6) q^(14) + t^(7) q^(16)'],
    dc.khovanov_odd_integral_polynomial: [
        '',
        'q^(2) + t^(2) q^(6) + t^(3) q^(8)',
        't^(-2) q^(-4) + t^(-1) q^(-2) + 1 + t q^(2) + t^(2) q^(4)',
        'q^(4) + t^(2) q^(8) + t^(3) q^(10) + t^(4) q^(12) + t^(5) q^(14)',
        'q^(2) + t q^(4) + 2 t^(2) q^(6) + t^(3) q^(8) + t^(4) q^(10) + t^(5) q^(12)',
        't^(-2) q^(-4) + t^(-1) q^(-2) + 2 + 2 t q^(2) + t^(2) q^(4) + t^(3) q^(6) + t^(4) q^(8)',
        't^(-2) q^(-2) + t^(-1) + 2 q^(2) + 2 t q^(4) + 2 t^(2) q^(6) + 2 t^(3) q^(8) + t^(4) q^(10)',
        't^(-3) q^(-6) + 2 t^(-2) q^(-4) + 2 t^(-1) q^(-2) + 3 + 2 t q^(2) + 2 t^(2) q^(4) + t^(3) q^(6)',
        'q^(6) + t^(2) q^(10) + t^(3) q^(12) + t^(4) q^(14) + t^(5) q^(16) + t^(6) q^(18) + t^(7) q^(20)',
        'q^(2) + t q^(4) + 2 t^(2) q^(6) + 2 t^(3) q^(8) + 2 t^(4) q^(10) + t^(5) q^(12) + t^(6) q^(14) + t^(7) q^(16)'],
    dc.khovanov_odd_rational_polynomial: [
        '',
        ' q^(2) + t^(2) q^(6) + t^(3) q^(8)',
        't^(-2) q^(-4) + t^(-1) q^(-2) + 1 + t q^(2) + t^(2) q^(4)',
        ' q^(4) + t^(2) q^(8) + t^(3) q^(10) + t^(4) q^(12) + t^(5) q^(14)',
        ' q^(2) + t q^(4) + 2 t^(2) q^(6) + t^(3) q^(8) + t^(4) q^(10) + t^(5) q^(12)',
        't^(-2) q^(-4) + t^(-1) q^(-2) + 2 + 2 t q^(2) + t^(2) q^(4) + t^(3) q^(6) + t^(4) q^(8)',
        't^(-2) q^(-2) + t^(-1) + 2 q^(2) + 2 t q^(4) + 2 t^(2) q^(6) + 2 t^(3) q^(8) + t^(4) q^(10)',
        't^(-3) q^(-6) + 2 t^(-2) q^(-4) + 2 t^(-1) q^(-2) + 3 + 2 t q^(2) + 2 t^(2) q^(4) + t^(3) q^(6)',
        ' q^(6) + t^(2) q^(10) + t^(3) q^(12) + t^(4) q^(14) + t^(5) q^(16) + t^(6) q^(18) + t^(7) q^(20)',
        ' q^(2) + t q^(4) + 2 t^(2) q^(6) + 2 t^(3) q^(8) + 2 t^(4) q^(10) + t^(5) q^(12) + t^(6) q^(14) + t^(7) q^(16)'],
    dc.khovanov_odd_mod2_polynomial: [
        '',
        ' q^(2) + t^(2) q^(6) + t^(3) q^(8)',
        't^(-2) q^(-4) + t^(-1) q^(-2) + 1 + t q^(2) + t^(2) q^(4)',
        ' q^(4) + t^(2) q^(8) + t^(3) q^(10) + t^(4) q^(12) + t^(5) q^(14)',
        ' q^(2) + t q^(4) + 2 t^(2) q^(6) + t^(3) q^(8) + t^(4) q^(10) + t^(5) q^(12)',
        't^(-2) q^(-4) + t^(-1) q^(-2) + 2 + 2 t q^(2) + t^(2) q^(4) + t^(3) q^(6) + t^(4) q^(8)',
        't^(-2) q^(-2) + t^(-1) + 2 q^(2) + 2 t q^(4) + 2 t^(2) q^(6) + 2 t^(3) q^(8) + t^(4) q^(10)',
        't^(-3) q^(-6) + 2 t^(-2) q^(-4) + 2 t^(-1) q^(-2) + 3 + 2 t q^(2) + 2 t^(2) q^(4) + t^(3) q^(6)',
        ' q^(6) + t^(2) q^(10) + t^(3) q^(12) + t^(4) q^(14) + t^(5) q^(16) + t^(6) q^(18) + t^(7) q^(20)',
        ' q^(2) + t q^(4) + 2 t^(2) q^(6) + 2 t^(3) q^(8) + 2 t^(4) q^(10) + t^(5) q^(12) + t^(6) q^(14) + t^(7) q^(16)']
}



# sage_setup: distribution = sagemath-graphs
# cython: binding=True
r"""
Comparability and permutation graphs

This module implements method related to :wikipedia:`Comparability_graph` and
:wikipedia:`Permutation_graph`, that is, for the moment, only recognition
algorithms.

Most of the information found here can also be found in [ST1994]_ or [Sha1997]_.

The following methods are implemented in this module

.. csv-table::
    :class: contentstable
    :widths: 30, 70
    :delim: |

    :meth:`~is_comparability_MILP` | Check whether the graph is a comparability graph (MILP)
    :meth:`~greedy_is_comparability` | Check whether the graph is a comparability graph (greedy algorithm)
    :meth:`~greedy_is_comparability_with_certificate` | Check whether the graph is a comparability graph and returns certificates (greedy algorithm)
    :meth:`~is_comparability` | Check whether the graph is a comparability graph
    :meth:`~is_permutation` | Check whether the graph is a permutation graph.
    :meth:`~is_transitive` | Check whether the digraph is transitive.

Author:

- Nathann Cohen 2012-04

Graph classes
-------------

**Comparability graphs**

A graph is a comparability graph if it can be obtained from a poset by adding an
edge between any two elements that are comparable. Co-comparability graph are
complements of such graphs, i.e. graphs built from a poset by adding an edge
between any two incomparable elements.

For more information on comparability graphs, see the
:wikipedia:`Comparability_graph`.

**Permutation graphs**

Definitions:

- A permutation `\pi = \pi_1\pi_2\dots\pi_n` defines a graph on `n` vertices
  such that `i\sim j` when `\pi` reverses `i` and `j` (i.e. when `i<j` and
  `\pi_j < \pi_i`. A graph is a permutation graph whenever it can be built
  through this construction.

- A graph is a permutation graph if it can be build from two parallel lines are
  the intersection graph of segments intersecting both lines.

- A graph is a permutation graph if it is both a comparability graph and a
  co-comparability graph.

For more information on permutation graphs, see the
:wikipedia:`Permutation_graph`.


Recognition algorithm for comparability graphs
----------------------------------------------

**Greedy algorithm**

This algorithm attempts to build a transitive orientation of a given graph `G`,
that is an orientation `D` such that for any directed `uv`-path of `D` there
exists in `D` an edge `uv`. This already determines a notion of equivalence
between some edges of `G` :

  In `G`, two edges `uv` and `uv'` (incident to a common vertex `u`) such that
  `vv'\not\in G` need necessarily be oriented *the same way* (that is that they
  should either both *leave* or both *enter* `u`). Indeed, if one enters `G`
  while the other leaves it, these two edges form a path of length two, which is
  not possible in any transitive orientation of `G` as `vv'\not\in G`.

Hence, we can say that in this case a *directed edge* `uv` is equivalent to a
*directed edge* `uv'` (to mean that if one belongs to the transitive
orientation, the other one must be present too) in the same way that `vu` is
equivalent to `v'u`. We can thus define equivalence classes on oriented edges,
to represent set of edges that imply each other. We can thus define `C^G_{uv}`
to be the equivalence class in `G` of the oriented edge `uv`.

Of course, if there exists a transitive orientation of a graph `G`, then no edge
`uv` implies its contrary `vu`, i.e. it is necessary to ensure that `\forall
uv\in G, vu\not\in C^G_{uv}`. The key result on which the greedy algorithm is
built is the following (see [ST1994]_):

  **Theorem** -- The following statements are equivalent :

  - `G` is a comparability graph
  - `\forall uv\in G, vu\not\in C^G_{uv}`
  - The edges of `G` can be partitioned into `B_1,...,B_k` where `B_i` is the
    equivalence class of some oriented edge in `G-B_1-\dots-B_{i-1}`

Hence, ensuring that a graph is a comparability graph can be done by checking
that no equivalence class is contradictory. Building the orientation, however,
requires to build equivalence classes step by step until an orientation has been
found for all of them.

**Mixed Integer Linear Program**

A MILP formulation is available to check the other methods for correction. It is
easily built :

  To each edge are associated two binary variables (one for each possible
  direction).  We then ensure that each triangle is transitively oriented, and
  that each pair of incident edges `uv, uv'` such that `vv'\not\in G` do not
  create a 2-path.

Here is the formulation:

.. MATH::

    \mbox{Maximize : }&\mbox{Nothing}\\
    \mbox{Such that : }&\\
    &\forall uv\in G\\
    &\cdot o_{uv}+o_{vu} = 1\\
    &\forall u\in G, \forall v,v'\in N(v)\text{ such that }vv'\not\in G\\
    &\cdot o_{uv} + o_{v'u} - o_{v'v} \leq 1\\
    &\cdot o_{uv'} + o_{vu} - o_{vv'} \leq 1\\
    &\forall u\in G, \forall v,v'\in N(v)\text{ such that }vv'\in G\\
    &\cdot o_{uv} + o_{v'u} \leq  1\\
    &\cdot o_{uv'} + o_{vu} \leq  1\\
    &o_{uv}\text{ is a binary variable}\\

.. NOTE::

  The MILP formulation is usually much slower than the greedy algorithm. This
  MILP has been implemented to check the results of the greedy algorithm that
  has been implemented to check the results of a faster algorithm which has not
  been implemented yet.

Certificates
------------

**Comparability graphs**

The *yes*-certificates that a graph is a comparability graphs are transitive
orientations of it. The *no*-certificates, on the other hand, are odd cycles of
such graph. These odd cycles have the property that around each vertex `v` of
the cycle its two incident edges must have the same orientation (toward `v`, or
outward `v`) in any transitive orientation of the graph. This is impossible
whenever the cycle has odd length. Explanations are given in the "Greedy
algorithm" part of the previous section.

**Permutation graphs**

Permutation graphs are precisely the intersection of comparability graphs and
co-comparability graphs. Hence, negative certificates are precisely negative
certificates of comparability or co-comparability. Positive certificates are a
pair of permutations that can be used through
:meth:`~sage.graphs.graph_generators.GraphGenerators.PermutationGraph` (whose
documentation says more about what these permutations represent).

Implementation details
----------------------

**Test that the equivalence classes are not self-contradictory**

This is done by a call to :meth:`Graph.is_bipartite`, and here is how :

   Around a vertex `u`, any two edges `uv, uv'` such that `vv'\not\in G` are
   equivalent. Hence, the equivalence class of edges around a vertex are
   precisely the connected components of the complement of the graph induced by
   the neighbors of `u`.

   In each equivalence class (around a given vertex `u`), the edges should all
   have the same orientation, i.e. all should go toward `u` at the same time, or
   leave it at the same time. To represent this, we create a graph with vertices
   for all equivalent classes around all vertices of `G`, and link `(v, C)` to
   `(u, C')` if `u\in C` and `v\in C'`.

   A bipartite coloring of this graph with colors 0 and 1 tells us that the
   edges of an equivalence class `C` around `u` should be directed toward `u` if
   `(u, C)` is colored with `0`, and outward if `(u, C)` is colored with `1`.

   If the graph is not bipartite, this is the proof that some equivalence class
   is self-contradictory !


.. NOTE::

    The greedy algorithm implemented here is just there to check the correction
    of more complicated ones, and it is reaaaaaaaaaaaalllly bad whenever you
    look at it with performance in mind.

Methods
-------
"""

# ****************************************************************************
#       Copyright (C) 2012 Nathann Cohen <nathann.cohen@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from cysignals.memory cimport sig_free
from sage.graphs.distances_all_pairs cimport c_distances_all_pairs
from copy import copy


#####################
# Greedy Algorithms #
#####################

def greedy_is_comparability(g, no_certificate=False, equivalence_class=False):
    r"""
    Check whether the graph is a comparability graph (greedy algorithm).

    This method only returns no-certificates.

    To understand how this method works, please consult the documentation of the
    :mod:`comparability module <sage.graphs.comparability>`.

    INPUT:

    - ``g`` -- a graph

    - ``no_certificate`` -- whether to return a *no*-certificate when the graph
      is not a comparability graph. This certificate is an odd cycle of edges,
      each of which implies the next. It is set to ``False`` by default.

    - ``equivalence_class`` -- whether to return an equivalence class
      if the graph is a comparability graph

    OUTPUT:

    - If the graph is a comparability graph and ``no_certificate = False``, this
      method returns ``True`` or ``(True, an_equivalence_class)`` according to
      the value of ``equivalence_class``.

    - If the graph is *not* a comparability graph, this method returns ``False``
      or ``(False, odd_cycle)`` according to the value of ``no_certificate``.

    EXAMPLES:

    The Petersen Graph is not transitively orientable::

        sage: from sage.graphs.comparability import greedy_is_comparability as is_comparability
        sage: g = graphs.PetersenGraph()
        sage: is_comparability(g)
        False
        sage: is_comparability(g, no_certificate=True)
        (False, [2, 1, 0, 4, 3, 2])

    But the Bull graph is::

        sage: g = graphs.BullGraph()
        sage: is_comparability(g)
        True

    TESTS:

    Check that the method is working even when vertices are of incomparable
    types::

        sage: from sage.graphs.comparability import greedy_is_comparability
        sage: G = Graph([('a', 1), (1, 2), (2, 3)])
        sage: greedy_is_comparability(G, equivalence_class=True)
        (True, [('a', 1), (2, 1), (2, 3)])
    """
    cdef int i, j

    # Each vertex can partition its neighbors into equivalence classes
    cdef dict equivalence_classes = {}
    for v in g:
        equivalence_classes[v] = g.subgraph(vertices=g.neighbors(v)).complement().connected_components(sort=False)

    # We build a graph h with one vertex per (vertex of g + equivalence class)
    from sage.graphs.graph import Graph
    h = Graph()
    h.add_vertices([(v, i) for v in g for i in range(len(equivalence_classes[v]))])

    # We add an edge between two vertices of h if they represent
    # opposed equivalence classes

    for u, v in g.edge_iterator(labels=False):

        for i, s in enumerate(equivalence_classes[v]):
            if u in s:
                break

        for j, s in enumerate(equivalence_classes[u]):
            if v in s:
                break

        h.add_edge((v, i), (u, j))

    # Is it a comparability graph ?

    cdef int isit
    isit, certif = h.is_bipartite(certificate=True)

    if isit:
        if equivalence_class:
            # We use a mapping between vertices and integers to deal with
            # vertices of different types
            int_to_vertex = list(g)
            vertex_to_int = {u: i for i, u in enumerate(int_to_vertex)}

            # Returning the largest equivalence class
            cc = max(h.connected_components(sort=False), key=len)

            edges = []
            for v, sid in cc:
                s = equivalence_classes[v][sid]

                # For each edge we pick the good orientations
                vi = vertex_to_int[v]
                if certif[v, sid] == 1:
                    edges.extend((vi, vertex_to_int[vv]) for vv in s)
                else:
                    edges.extend((vertex_to_int[vv], vi) for vv in s)

            # We return the value but take care of removing edges that were
            # added twice.
            edges = [(int_to_vertex[u], int_to_vertex[v]) for u, v in sorted(set(edges))]
            return True, edges

        return True

    if no_certificate:
        cycle = [v for v, _ in certif]
        cycle.append(cycle[0])
        return False, cycle
    return False


def greedy_is_comparability_with_certificate(g, certificate=False):
    r"""
    Check whether the graph is a comparability graph and returns
    certificates (greedy algorithm).

    This method can return certificates of both *yes* and *no* answers.

    To understand how this method works, please consult the documentation of the
    :mod:`comparability module <sage.graphs.comparability>`.

    INPUT:

    - ``g`` -- a graph

    - ``certificate`` -- boolean; whether to return a
      certificate. *Yes*-answers the certificate is a transitive orientation of
      `G`, and a *no* certificates is an odd cycle of sequentially forcing
      edges.

    EXAMPLES:

    The 5-cycle or the Petersen Graph are not transitively orientable::

        sage: from sage.graphs.comparability import greedy_is_comparability_with_certificate as is_comparability
        sage: is_comparability(graphs.CycleGraph(5), certificate=True)
        (False, [2, 1, 0, 4, 3, 2])
        sage: g = graphs.PetersenGraph()
        sage: is_comparability(g)
        False
        sage: is_comparability(g, certificate=True)
        (False, [2, 1, 0, 4, 3, 2])

    But the Bull graph is::

        sage: g = graphs.BullGraph()
        sage: is_comparability(g)
        True
        sage: is_comparability(g, certificate = True)
        (True, Digraph on 5 vertices)
        sage: is_comparability(g, certificate = True)[1].is_transitive()
        True

    TESTS:

    Check that the method is working even when vertices are of incomparable
    types::

        sage: from sage.graphs.comparability import greedy_is_comparability_with_certificate
        sage: G = Graph([('a', 1), (1, 2), (2, 3)])
        sage: greedy_is_comparability_with_certificate(G, certificate=True)
        (True, Digraph on 4 vertices)
    """
    isit, certif = greedy_is_comparability(g, no_certificate=True, equivalence_class=True)
    if not isit:
        if certificate:
            return False, certif
        return False

    elif not certificate:
        return True

    gg = copy(g)
    from sage.graphs.digraph import DiGraph
    h = DiGraph()
    h.add_vertices(gg)

    for u, v in certif:
        gg.delete_edge(u, v)
        h.add_edge(u, v)

    # While there are some edges left to be oriented
    while gg.size():

        # We take an equivalence class and orient it
        isit, certif = greedy_is_comparability(gg, no_certificate=True, equivalence_class=True)

        # Then remove it from the former graph
        for u, v in certif:
            gg.delete_edge(u, v)
            h.add_edge(u, v)

    return True, h


###################
# Integer Program #
###################

def is_comparability_MILP(g, certificate=False, solver=None, verbose=0):
    r"""
    Check whether the graph is a comparability graph (MILP).

    INPUT:

    - ``g`` -- a graph

    - ``certificate`` -- boolean (default: ``False``); whether to return a
      certificate for yes instances. This method cannot return negative
      certificates.

    - ``solver`` -- string (default: ``None``); specifies a Mixed Integer Linear
      Programming (MILP) solver to be used. If set to ``None``, the default one
      is used. For more information on MILP solvers and which default solver is
      used, see the method :meth:`solve
      <sage.numerical.mip.MixedIntegerLinearProgram.solve>` of the class
      :class:`MixedIntegerLinearProgram
      <sage.numerical.mip.MixedIntegerLinearProgram>`.

    - ``verbose`` -- integer (default: 0); sets the level of verbosity. Set
      to 0 by default, which means quiet.

    EXAMPLES:

    The 5-cycle or the Petersen Graph are not transitively orientable::

        sage: from sage.graphs.comparability import is_comparability_MILP as is_comparability
        sage: is_comparability(graphs.CycleGraph(5), certificate=True)                  # needs sage.numerical.mip
        (False, None)
        sage: g = graphs.PetersenGraph()
        sage: is_comparability(g, certificate=True)                                     # needs sage.numerical.mip
        (False, None)

    But the Bull graph is::

        sage: g = graphs.BullGraph()
        sage: is_comparability(g)                                                       # needs sage.numerical.mip
        True
        sage: is_comparability(g, certificate=True)                                     # needs sage.numerical.mip
        (True, Digraph on 5 vertices)
        sage: is_comparability(g, certificate=True)[1].is_transitive()                  # needs sage.numerical.mip
        True
    """
    from sage.numerical.mip import MixedIntegerLinearProgram, MIPSolverException
    p = MixedIntegerLinearProgram(solver=solver)
    o = p.new_variable(binary=True)

    for u, v in g.edge_iterator(labels=False):
        p.add_constraint(o[u, v] + o[v, u] == 1)

    from itertools import combinations
    for u in g:
        for v, vv in combinations(g.neighbors(u), 2):

            # If there is an edge between v and vv, we must be sure it is in the
            # good direction when v-u-vv is a directed path
            if g.has_edge(v, vv):
                p.add_constraint(o[u, v] + o[vv, u] - o[vv, v] <= 1)
                p.add_constraint(o[u, vv] + o[v, u] - o[v, vv] <= 1)

            # If there is no edge, there are only two orientations possible (see
            # the module's documentation about edges which imply each other)
            else:
                p.add_constraint(o[u, v] + o[vv, u] <= 1)
                p.add_constraint(o[u, vv] + o[v, u] <= 1)

    try:
        p.solve(log=verbose)
        if not certificate:
            return True

        # Building the transitive orientation
        from sage.graphs.digraph import DiGraph
        d = DiGraph()
        d.add_vertices(g)

        tol = 0 if p.base_ring().is_exact() else 1e-6
        o = p.get_values(o, convert=True, tolerance=tol)
        for u, v in g.edge_iterator(labels=False):
            if o[u, v]:
                d.add_edge(u, v)
            else:
                d.add_edge(v, u)

        return True, d

    except MIPSolverException:
        if certificate:
            return False, None
        return False


###############
# Empty shell #
###############

def is_comparability(g, algorithm='greedy', certificate=False, check=True,
                     solver=None, verbose=0):
    r"""
    Check whether the graph is a comparability graph.

    INPUT:

    - ``g`` -- a graph

    - ``algorithm`` -- string (default: ``'greedy'``); choose the implementation
      used to do the test

      - ``'greedy'`` -- a greedy algorithm (see the documentation of the
        :mod:`comparability module <sage.graphs.comparability>`)

      - ``'MILP'`` -- a Mixed Integer Linear Program formulation of the
        problem. Beware, for this implementation is unable to return negative
        certificates ! When ``certificate = True``, negative certificates are
        always equal to ``None``. ``True`` certificates are valid, though.

    - ``certificate`` -- boolean (default: ``False``); whether to return a
      certificate. *Yes*-answers the certificate is a transitive orientation of
      `G`, and a *no* certificates is an odd cycle of sequentially forcing
      edges.

    - ``check`` -- boolean (default: ``True``); whether to check that the
      yes-certificates are indeed transitive. As it is very quick
      compared to the rest of the operation, it is enabled by default.

    - ``solver`` -- string (default: ``None``); specifies a Mixed Integer Linear
      Programming (MILP) solver to be used. If set to ``None``, the default one
      is used. For more information on MILP solvers and which default solver is
      used, see the method :meth:`solve
      <sage.numerical.mip.MixedIntegerLinearProgram.solve>` of the class
      :class:`MixedIntegerLinearProgram
      <sage.numerical.mip.MixedIntegerLinearProgram>`.

    - ``verbose`` -- integer (default: 0); sets the level of verbosity. Set
      to 0 by default, which means quiet.

    EXAMPLES::

        sage: from sage.graphs.comparability import is_comparability
        sage: g = graphs.PetersenGraph()
        sage: is_comparability(g)
        False
        sage: is_comparability(graphs.CompleteGraph(5), certificate=True)
        (True, Digraph on 5 vertices)

    TESTS:

    Let us ensure that no exception is raised when we go over all small graphs::

        sage: from sage.graphs.comparability import is_comparability
        sage: [len([g for g in graphs(i) if is_comparability(g, certificate=True)[0]]) for i in range(7)]   # needs nauty
        [1, 1, 2, 4, 11, 33, 144]
    """
    g._scream_if_not_simple()
    if not g.size():
        if certificate:
            from sage.graphs.digraph import DiGraph
            return True, DiGraph(g)
        return True

    if algorithm == "greedy":
        comparability_test = greedy_is_comparability_with_certificate(g, certificate=certificate)
    elif algorithm == "MILP":
        comparability_test = is_comparability_MILP(g, certificate=certificate,
                                                   solver=solver, verbose=verbose)

    if not certificate:
        return comparability_test

    # Checking that the orientation found is indeed transitive. No
    # reason why it should not, but no reason why we should not check
    # anyway :-p
    isit, certif = comparability_test

    if check and isit and (not certif.is_transitive()):
        raise ValueError("Looks like there is a bug somewhere. The "
                         "algorithm thinks that the orientation is "
                         "transitive, but we just checked and it is not. "
                         "Please report the bug on sage-devel, and give "
                         "us the graph that made this method fail !")

    return isit, certif


def is_permutation(g, algorithm='greedy', certificate=False, check=True,
                   solver=None, verbose=0):
    r"""
    Check whether the graph is a permutation graph.

    For more information on permutation graphs, refer to the documentation of
    the :mod:`comparability module <sage.graphs.comparability>`.

    INPUT:

    - ``g`` -- a graph

    - ``algorithm`` -- string (default: ``'greedy'``); choose the implementation
      used for the subcalls to :meth:`is_comparability`

      - ``'greedy'`` -- a greedy algorithm (see the documentation of the
        :mod:`comparability module <sage.graphs.comparability>`)

      - ``'MILP'`` -- a Mixed Integer Linear Program formulation of the
        problem. Beware, for this implementation is unable to return negative
        certificates ! When ``certificate = True``, negative certificates are
        always equal to ``None``. ``True`` certificates are valid, though.

    - ``certificate`` -- boolean (default: ``False``); whether to return a
      certificate for the answer given. For ``True`` answers the certificate is
      a permutation, for ``False`` answers it is a no-certificate for the test
      of comparability or co-comparability.

    - ``check`` -- boolean (default: ``True``); whether to check that the
      permutations returned indeed create the expected Permutation graph. Pretty
      cheap compared to the rest, hence a good investment. It is enabled by
      default.

    - ``solver`` -- string (default: ``None``); specifies a Mixed Integer Linear
      Programming (MILP) solver to be used. If set to ``None``, the default one
      is used. For more information on MILP solvers and which default solver is
      used, see the method :meth:`solve
      <sage.numerical.mip.MixedIntegerLinearProgram.solve>` of the class
      :class:`MixedIntegerLinearProgram
      <sage.numerical.mip.MixedIntegerLinearProgram>`.

    - ``verbose`` -- integer (default: 0); sets the level of verbosity. Set
      to 0 by default, which means quiet.

    .. NOTE::

        As the ``True`` certificate is a :class:`Permutation` object, the
        segment intersection model of the permutation graph can be visualized
        through a call to :meth:`Permutation.show
        <sage.combinat.permutation.Permutation.show>`.

    EXAMPLES:

    A permutation realizing the bull graph::

        sage: from sage.graphs.comparability import is_permutation
        sage: g = graphs.BullGraph()
        sage: _ , certif = is_permutation(g, certificate=True)
        sage: h = graphs.PermutationGraph(*certif)
        sage: h.is_isomorphic(g)
        True

    Plotting the realization as an intersection graph of segments::

        sage: true, perm = is_permutation(g, certificate=True)
        sage: p1 = Permutation([nn+1 for nn in perm[0]])
        sage: p2 = Permutation([nn+1 for nn in perm[1]])
        sage: p = p2 * p1.inverse()
        sage: p.show(representation='braid')                                            # needs sage.plot

    TESTS:

    Trying random permutations, first with the greedy algorithm::

        sage: from sage.graphs.comparability import is_permutation
        sage: for i in range(20):
        ....:     p = Permutations(10).random_element()
        ....:     g1 = graphs.PermutationGraph(p)
        ....:     isit, certif = is_permutation(g1, certificate=True)
        ....:     if not isit:
        ....:        print("Something is wrong here !!")
        ....:        break
        ....:     g2 = graphs.PermutationGraph(*certif)
        ....:     if not g1.is_isomorphic(g2):
        ....:        print("Something is wrong here !!")
        ....:        break

    Then with MILP::

        sage: from sage.graphs.comparability import is_permutation
        sage: for i in range(20):                                                       # needs sage.numerical.mip
        ....:     p = Permutations(10).random_element()
        ....:     g1 = graphs.PermutationGraph(p)
        ....:     isit, certif = is_permutation(g1, algorithm='MILP', certificate=True)
        ....:     if not isit:
        ....:        print("Something is wrong here !!")
        ....:        break
        ....:     g2 = graphs.PermutationGraph(*certif)
        ....:     if not g1.is_isomorphic(g2):
        ....:        print("Something is wrong here !!")
        ....:        break
    """
    if not certificate:
        # No certificate... A piece of cake
        return (is_comparability(g, algorithm=algorithm, solver=solver, verbose=verbose) and
                is_comparability(g.complement(), algorithm=algorithm, solver=solver, verbose=verbose))

    # First poset, we stop if it fails
    isit, certif = is_comparability(g, algorithm=algorithm, certificate=True,
                                    solver=solver, verbose=verbose)
    if not isit:
        return False, certif

    # Second poset
    isit, co_certif = is_comparability(g.complement(), algorithm=algorithm, certificate=True,
                                       solver=solver, verbose=verbose)
    if not isit:
        return False, co_certif

    # Building the two orderings
    tmp = list(co_certif.edges(labels=False, sort=False))
    for u, v in certif.edge_iterator(labels=False):
        co_certif.add_edge(v, u)
    certif.add_edges(tmp)

    ordering = certif.topological_sort()
    co_ordering = co_certif.topological_sort()

    # Try to build the Permutation graph from the permutations, just to make
    # sure nothing weird happened !
    if check:
        from sage.graphs.graph_generators import GraphGenerators
        pg = GraphGenerators().PermutationGraph(ordering, co_ordering)
        if not pg.is_isomorphic(g):
            raise ValueError("There is a mistake somewhere ! It looks like "
                             "the Permutation Graph model computed does "
                             "not match the input graph !")

    return True, (ordering, co_ordering)


def is_transitive(g, certificate=False):
    r"""
    Test whether the digraph is transitive.

    A digraph is transitive if for any pair of vertices `u,v\in G` linked by a
    `uv`-path the edge `uv` belongs to `G`.

    INPUT:

    - ``certificate`` -- whether to return a certificate for negative answers

      - If ``certificate = False`` (default), this method returns ``True`` or
        ``False`` according to the graph.

      - If ``certificate = True``, this method either returns ``True`` answers
        or yield a pair of vertices `uv` such that there exists a `uv`-path in
        `G` but `uv\not\in G`.

    EXAMPLES::

        sage: digraphs.Circuit(4).is_transitive()
        False
        sage: digraphs.Circuit(4).is_transitive(certificate=True)
        (0, 2)
        sage: digraphs.RandomDirectedGNP(30,.2).is_transitive()
        False
        sage: D = digraphs.DeBruijn(5, 2)                                               # needs sage.combinat
        sage: D.is_transitive()                                                         # needs sage.combinat
        False
        sage: cert = D.is_transitive(certificate=True)                                  # needs sage.combinat
        sage: D.has_edge(*cert)                                                         # needs sage.combinat
        False
        sage: bool(D.shortest_path(*cert))                                              # needs sage.combinat
        True
        sage: digraphs.RandomDirectedGNP(20,.2).transitive_closure().is_transitive()    # needs networkx
        True
    """
    cdef int n = g.order()

    if n <= 2:
        return True

    cdef list int_to_vertex = list(g)
    cdef unsigned short * distances = c_distances_all_pairs(g, vertex_list=int_to_vertex)
    cdef unsigned short * c_distances = distances

    cdef int i, j

    # Only 3 distances can appear in the matrix of all distances : 0, 1, and
    # infinity. Anything else is a proof of nontransitivity !

    for j in range(n):
        for i in range(n):
            if c_distances[i] != <unsigned short> -1 and c_distances[i] > 1:
                sig_free(distances)
                if certificate:

                    return int_to_vertex[j], int_to_vertex[i]
                else:
                    return False

        c_distances += n

    sig_free(distances)
    return True



# sage_setup: distribution = sagemath-graphs
# sage.doctest: needs sage.rings.finite_rings
r"""
Group-divisible designs (GDD)

This module gathers everything related to Group-Divisible Designs. The
constructions defined here can be accessed through ``designs.<tab>``::

    sage: designs.group_divisible_design(14,{4},{2})
    Group Divisible Design on 14 points of type 2^7

The main function implemented here is :meth:`group_divisible_design` (which
calls all others) and the main class is :class:`GroupDivisibleDesign`. The
following functions are available:

.. csv-table::
    :class: contentstable
    :widths: 30, 70
    :delim: |

    :func:`group_divisible_design` | Return a `(v,K,G)`-Group Divisible Design.
    :func:`GDD_4_2` | Return a `(2q,\{4\},\{2\})`-GDD for `q` a prime power with `q\equiv 1\pmod{6}`.

Functions
---------
"""

# ****************************************************************************
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.arith.misc import is_prime_power
from sage.misc.unknown import Unknown
from .incidence_structures import IncidenceStructure


def group_divisible_design(v, K, G, existence=False, check=False):
    r"""
    Return a `(v,K,G)`-Group Divisible Design.

    A `(v,K,G)`-GDD is a pair `(\mathcal G, \mathcal B)` where:

    - `\mathcal G` is a partition of `X=\bigcup \mathcal G` where `|X|=v`

    - `\forall S\in \mathcal G, |S| \in G`

    - `\forall S\in \mathcal B, |S| \in K`

    - `\mathcal G\cup \mathcal B` is a `(v,K\cup G)`-PBD

    For more information, see the documentation of
    :class:`~sage.combinat.designs.incidence_structures.GroupDivisibleDesign` or
    :class:`~sage.combinat.designs.bibd.PairwiseBalancedDesign`.

    INPUT:

    - ``v`` -- integer

    - ``K``, ``G`` -- sets of integers

    - ``existence`` -- boolean; instead of building the design, return:

        - ``True`` -- meaning that Sage knows how to build the design

        - ``Unknown`` -- meaning that Sage does not know how to build the
          design, but that the design may exist (see :mod:`sage.misc.unknown`)

        - ``False`` -- meaning that the design does not exist

    - ``check`` -- boolean (default: ``True``); whether to check that output is
      correct before returning it. As this is expected to be useless, you may
      want to disable it whenever you want speed.

    .. NOTE::

        The GDD returned by this function are defined on ``range(v)``, and its
        groups are sets of consecutive integers.

    EXAMPLES::

        sage: designs.group_divisible_design(14,{4},{2})
        Group Divisible Design on 14 points of type 2^7
    """
    G = list(set(G))
    K = list(set(K))

    blocks = None

    # from a (v+1,k,1)-BIBD
    if len(G) == 1 == len(K) and G[0] + 1 in K:
        from .bibd import balanced_incomplete_block_design
        k = K[0]
        if existence:
            return balanced_incomplete_block_design(v + 1, k, existence=True)
        BIBD = balanced_incomplete_block_design(v + 1, k)
        groups = [[x for x in S if x != v] for S in BIBD if v in S]
        d = {p: i for i, p in enumerate(sum(groups, []))}
        d[v] = v
        BIBD.relabel(d)
        groups = [list(range((k - 1) * i, (k - 1) * (i + 1)))
                  for i in range(v // (k - 1))]
        blocks = [S for S in BIBD if v not in S]

    # (v,{4},{2})-GDD
    elif (v % 2 == 0 and K == [4] and
          G == [2] and GDD_4_2(v // 2, existence=True)):
        if existence:
            return True
        return GDD_4_2(v // 2, check=check)

    # From a TD(k,g)
    elif len(G) == 1 == len(K) and K[0] * G[0] == v:
        from .orthogonal_arrays import transversal_design
        return transversal_design(k=K[0], n=G[0], existence=existence)

    if blocks:
        return GroupDivisibleDesign(v,
                                    groups=groups,
                                    blocks=blocks,
                                    G=G,
                                    K=K,
                                    check=check,
                                    copy=True)

    if existence:
        return Unknown
    raise NotImplementedError


def GDD_4_2(q, existence=False, check=True):
    r"""
    Return a `(2q,\{4\},\{2\})`-GDD for `q` a prime power with `q\equiv 1\pmod{6}`.

    This method implements Lemma VII.5.17 from [BJL99] (p.495).

    INPUT:

    - ``q`` -- integer

    - ``existence`` -- boolean; instead of building the design, return:

        - ``True`` -- meaning that Sage knows how to build the design

        - ``Unknown`` -- meaning that Sage does not know how to build the
          design, but that the design may exist (see :mod:`sage.misc.unknown`)

        - ``False`` -- meaning that the design does not exist

    - ``check`` -- boolean (default: ``True``); whether to check that output is
      correct before returning it. As this is expected to be useless, you may
      want to disable it whenever you want speed.

    EXAMPLES::

        sage: from sage.combinat.designs.group_divisible_designs import GDD_4_2
        sage: GDD_4_2(7,existence=True)
        True
        sage: GDD_4_2(7)
        Group Divisible Design on 14 points of type 2^7
        sage: GDD_4_2(8,existence=True)
        Unknown
        sage: GDD_4_2(8)
        Traceback (most recent call last):
        ...
        NotImplementedError
    """
    if q <= 1 or q % 6 != 1 or not is_prime_power(q):
        if existence:
            return Unknown
        raise NotImplementedError
    if existence:
        return True

    from sage.rings.finite_rings.finite_field_constructor import FiniteField
    G = FiniteField(q, 'x')
    w = G.primitive_element()
    e = w**((q - 1) // 3)

    # A first parallel class is defined. G acts on it, which yields all others.
    first_class = [[(0, 0), (1, w**i), (1, e * w**i), (1, e * e * w**i)]
                   for i in range((q - 1) // 6)]

    label = {p: i for i, p in enumerate(G)}
    classes = [[[2 * label[x[1] + g] + (x[0] + j) % 2 for x in S]
                for S in first_class]
               for g in G for j in range(2)]

    return GroupDivisibleDesign(2 * q,
        groups=[[i, i + 1] for i in range(0, 2 * q, 2)],
        blocks=sum(classes, []),
        K=[4],
        G=[2],
        check=check,
        copy=False)


class GroupDivisibleDesign(IncidenceStructure):
    r"""
    Group Divisible Design (GDD).

    Let `K` and `G` be sets of positive integers and let `\lambda` be a positive
    integer. A Group Divisible Design of index `\lambda` and order `v` is a
    triple `(V,\mathcal G,\mathcal B)` where:

    - `V` is a set of cardinality `v`

    - `\mathcal G` is a partition of `V` into groups whose size belongs to `G`

    - `\mathcal B` is a family of subsets of `V` whose size belongs to `K` such
      that any two points `p_1,p_2\in V` from different groups appear
      simultaneously in exactly `\lambda` elements of `\mathcal B`. Besides, a
      group and a block intersect on at most one point.

    If `K=\{k_1,...,k_l\}` and `G` has exactly `m_i` groups of cardinality `k_i`
    then `G` is said to have type `k_1^{m_1}...k_l^{m_l}`.

    INPUT:

    - ``points`` -- the underlying set. If ``points`` is an integer `v`, then
      the set is considered to be `\{0, ..., v-1\}`

    - ``groups`` -- the groups of the design. Set to ``None`` for an automatic
      guess (this triggers ``check=True`` and can thus cost some time)

    - ``blocks`` -- collection of blocks

    - ``G`` -- list of integers of which the sizes of the groups must be
      elements. Set to ``None`` (automatic guess) by default

    - ``K`` -- list of integers of which the sizes of the blocks must be
      elements. Set to ``None`` (automatic guess) by default

    - ``lambd`` -- integer (default: `1`); value of `\lambda`

    - ``check`` -- boolean (default: ``True``); whether to check that the design is indeed a `GDD`
      with the right parameters

    - ``copy`` -- (use with caution) if set to ``False`` then ``blocks`` must be
      a list of lists of integers. The list will not be copied but will be
      modified in place (each block is sorted, and the whole list is
      sorted). Your ``blocks`` object will become the instance's internal data.

    EXAMPLES::

        sage: from sage.combinat.designs.group_divisible_designs import GroupDivisibleDesign
        sage: TD = designs.transversal_design(4,10)
        sage: groups = [list(range(i*10,(i+1)*10)) for i in range(4)]
        sage: GDD = GroupDivisibleDesign(40,groups,TD); GDD
        Group Divisible Design on 40 points of type 10^4

    With unspecified groups::

        sage: # needs sage.schemes
        sage: D = designs.transversal_design(4,3).relabel(list('abcdefghiklm'),inplace=False).blocks()
        sage: GDD = GroupDivisibleDesign('abcdefghiklm',None,D)
        sage: sorted(GDD.groups())
        [['a', 'b', 'c'], ['d', 'e', 'f'], ['g', 'h', 'i'], ['k', 'l', 'm']]
    """
    def __init__(self, points, groups, blocks, G=None, K=None, lambd=1,
                 check=True, copy=True, **kwds):
        r"""
        Constructor function.

        EXAMPLES::

            sage: from sage.combinat.designs.group_divisible_designs import GroupDivisibleDesign
            sage: TD = designs.transversal_design(4,10)
            sage: groups = [list(range(i*10,(i+1)*10)) for i in range(4)]
            sage: GDD = GroupDivisibleDesign(40,groups,TD); GDD
            Group Divisible Design on 40 points of type 10^4
        """
        from .designs_pyx import is_group_divisible_design

        self._lambd = lambd

        IncidenceStructure.__init__(self,
                                    points,
                                    blocks,
                                    copy=copy,
                                    check=False,
                                    **kwds)

        if groups is None or (copy is False and self._point_to_index is None):
            self._groups = groups
        elif self._point_to_index is None:
            self._groups = [g[:] for g in groups]
        else:
            self._groups = [[self._point_to_index[x] for x in g] for g in groups]

        if check or groups is None:
            is_gdd = is_group_divisible_design(self._groups, self._blocks,
                                               self.num_points(), G, K,
                                               lambd, verbose=1)
            assert is_gdd
            if groups is None:
                self._groups = is_gdd[1]

    def groups(self):
        r"""
        Return the groups of the Group-Divisible Design.

        EXAMPLES::

            sage: from sage.combinat.designs.group_divisible_designs import GroupDivisibleDesign
            sage: TD = designs.transversal_design(4,10)
            sage: groups = [list(range(i*10,(i+1)*10)) for i in range(4)]
            sage: GDD = GroupDivisibleDesign(40,groups,TD); GDD
            Group Divisible Design on 40 points of type 10^4
            sage: GDD.groups()
            [[0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
             [10, 11, 12, 13, 14, 15, 16, 17, 18, 19],
             [20, 21, 22, 23, 24, 25, 26, 27, 28, 29],
             [30, 31, 32, 33, 34, 35, 36, 37, 38, 39]]

        TESTS:

        Non-integer ground set::

            sage: # needs sage.schemes
            sage: TD = designs.transversal_design(5,5)
            sage: TD.relabel({i: chr(97+i) for i in range(25)})
            sage: TD.groups()
            [['a', 'b', 'c', 'd', 'e'],
             ['f', 'g', 'h', 'i', 'j'],
             ['k', 'l', 'm', 'n', 'o'],
             ['p', 'q', 'r', 's', 't'],
             ['u', 'v', 'w', 'x', 'y']]
        """
        if self._point_to_index is None:
            return [list(g) for g in self._groups]
        else:
            return [[self._points[i] for i in g] for g in self._groups]

    def __repr__(self):
        r"""
        Return a string that describes ``self``.

        EXAMPLES::

            sage: from sage.combinat.designs.group_divisible_designs import GroupDivisibleDesign
            sage: TD = designs.transversal_design(4,10)
            sage: groups = [list(range(i*10,(i+1)*10)) for i in range(4)]
            sage: GDD = GroupDivisibleDesign(40,groups,TD); GDD
            Group Divisible Design on 40 points of type 10^4
        """
        group_sizes = [len(g) for g in self._groups]

        gdd_type = ("{}^{}".format(s, group_sizes.count(s))
                    for s in sorted(set(group_sizes)))
        gdd_type = ".".join(gdd_type)

        if not gdd_type:
            gdd_type = "1^0"

        v = self.num_points()

        return "Group Divisible Design on {} points of type {}".format(v, gdd_type)



# sage_setup: distribution = sagemath-graphs
# sage.doctest: needs sage.combinat
r"""
Cographs

A cograph is a `P_4`-free graph, that is a graph without induced path of order
4. Any cograph may be constructed, starting from the single vertex graph, by a
sequence of join and disjoint union operations. See the :wikipedia:`Cograph` for
more details on this graph class, and :oeis:`A000084` to know the number of
cographs of order `n \geq 1`.

This module implements the following methods concerning cographs:

.. csv-table::
    :class: contentstable
    :widths: 30, 70
    :delim: |

    :meth:`cographs` | Return an iterator over the cographs of order `n`.

Methods
-------
"""
# ****************************************************************************
#       Copyright (C) 2017-2024 Marianna Spyrakou
#                          2024 David Coudert <david.coudert@inria.fr>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.misc.lazy_import import lazy_import

lazy_import('sage.combinat.partitions', 'AccelAsc_next')


class CoTree:
    r"""
    Generic cotree node.

    This data structure is used for the generation of cographs in
    :meth:`cographs`.
    """
    def __init__(self, name='root'):
        r"""
        Initialize a cotree.

        INPUT:

        - ``name`` -- either an operation ('U' or 'J') or the size of the
          subtree rooted at this node

        EXAMPLES::

            sage: from sage.graphs.cographs import CoTree
            sage: CoTree(1)
            ( 1 )
            sage: CoTree()
            ( root )
        """
        self.name = name
        self.children = []
        self.info = None
        self.parent = None

    def __str__(self):
        r"""
        Return a string representation of ``self``.

        The subtree of a node is inside brackets and starts with the operation
        to perform between the children ('J' for join or 'U' for disjoint
        union).

        EXAMPLES::

            sage: from sage.graphs.cographs import CoTree
            sage: CoTree(1)
            ( 1 )
            sage: CoTree('J')
            [ J ]
            sage: next(graphs.cographs(4, as_graph=False))  # indirect doctest
            [ J ( 0 ) ( 1 ) ( 2 ) ( 3 ) ]
        """
        first = '[' if self.name in ['J', 'U'] else '('
        last = ' ]' if self.name in ['J', 'U'] else ' )'
        s = f"{first} {self.name}"
        for child in self.children:
            s += f' {child}'
        s += last
        return s

    __repr__ = __str__

    def add_child(self, node):
        r"""
        Add cotree ``node`` in the list of children of ``self``.

        INPUT:

        - ``node`` -- a CoTree

        EXAMPLES::

            sage: from sage.graphs.cographs import CoTree
            sage: T = CoTree('J')
            sage: T.add_child(CoTree(1))
            sage: T
            [ J ( 1 ) ]
        """
        assert isinstance(node, CoTree)
        self.children.append(node)
        node.parent = self

    def copy_tree(self, T):
        r"""
        Make `T` a copy of ``self``.

        INPUT:

        - ``T`` -- a CoTree

        EXAMPLES::

            sage: from sage.graphs.cographs import CoTree
            sage: T = CoTree('J')
            sage: T.add_child(CoTree(1))
            sage: T.add_child(CoTree(2))
            sage: T
            [ J ( 1 ) ( 2 ) ]
            sage: B =  CoTree('U')
            sage: T.copy_tree(B)
            sage: B
            [ U ( 1 ) ( 2 ) ]
        """
        for i, child in enumerate(self.children):
            T.add_child(CoTree(child.name))
            child.copy_tree(T.children[i])

    def reset_info(self):
        r"""
        Reset parameter ``info`` from all nodes of ``self``.

        EXAMPLES::

            sage: from sage.graphs.cographs import CoTree
            sage: T = CoTree(1)
            sage: B = CoTree(2)
            sage: T.add_child(B)
            sage: C = CoTree(3)
            sage: B.add_child(C)
            sage: C.info = 'info'
            sage: T.reset_info()
            sage: C.info is None
            True
        """
        for child in self.children:
            child.reset_info()
        self.info = None


def rebuild_node(u, P):
    r"""
    Replace the subtree rooted at `u` by a subtree induced by partition `P`.

    This is a helper method to method :meth:`cographs`.

    INPUT:

    - ``u`` -- a ``CoTree``

    - ``P`` -- a partition encoding the new structure of the children of `u`

    EXAMPLES::

        sage: next(graphs.cographs(3, as_graph=True)).vertices()  # indirect doctest
        [0, 1, 2]
    """
    if P is None:
        print('P is None')
    u.children = []  # delete the subtree rooted at u
    for value in P:
        this_child = CoTree(value)
        u.add_child(this_child)
        if value > 1:
            for j in range(value):
                this_child.add_child(CoTree(1))


def find_pivot(T):
    r"""
    Search for a pivot node in `T`.

    This is a helper method to method :meth:`cographs`.

    A node in `T` is a ``pivot`` if it is not a leaf, it does not induce a
    maximum partition and it is the first such node in the inverse postorder
    traversal.

    INPUT:

    - ``T`` -- a ``CoTree``

    EXAMPLES::

        sage: next(graphs.cographs(3, as_graph=True)).vertices()  # indirect doctest
        [0, 1, 2]
    """
    for child in reversed(T.children):
        pivot = find_pivot(child)
        if pivot is not None:
            return pivot

    # Check if T is a pivot
    i = T.name
    if (i != 1 and ((i//2 != T.children[0].name) or
                    (i//2 + i % 2 != T.children[1].name))):
        T.info = 'p'  # pivot mark
        return T
    return None


def next_tree(T):
    r"""
    Check if there is another tree after `T`.

    This is a helper method to method :meth:`cographs`.

    This methods returns ``True`` if there is a tree after `T`, and if so it
    modifies the input tree `T` that becomes the next tree.

    INPUT:

    - ``T`` -- a ``CoTree``

    EXAMPLES::

        sage: next(graphs.cographs(3, as_graph=True)).vertices()  # indirect doctest
        [0, 1, 2]
    """
    pivot = find_pivot(T)
    if pivot is None:
        return False

    # Find the next partition induced by the subtree pivot.
    # Partitions are represented in ascending order (i.e., `P_i \leq P_{i+1}`)
    # and we search for the next partition in lexicographic order.
    partition = [c.name for c in pivot.children]
    P = AccelAsc_next(partition)
    # and rebuild the subtree of pivot accordingly
    rebuild_node(pivot, P)

    x = pivot
    while x.parent is not None:
        ancestor = x.parent
        # Modify the bigger siblings of x (i.e., the siblings placed after
        # x in the list of children of ancestor)
        is_bigger_sibling = False
        for y, this_child in enumerate(ancestor.children):
            if this_child.info == 'p':
                is_bigger_sibling = True
            elif is_bigger_sibling:  # true only for bigger siblings of x
                if x.name == this_child.name:
                    temp = CoTree(x.name)
                    x.copy_tree(temp)
                    ancestor.children[y] = temp  # copy subtree T(x) in T(y)
                    temp.parent = ancestor
                else:
                    P = [1] * ancestor.children[y].name
                    rebuild_node(ancestor.children[y], P)

        # Make the parent of x (if any) the new pivot
        x.info = None  # reset the pivot mark
        ancestor.info = 'p'  # parent gets the pivot mark
        x = ancestor

    return True


def cographs(n, as_graph=True, immutable=False):
    r"""
    Return an iterator over the cographs of order `n`.

    A cograph is a `P_4`-free graph, that is a graph without induced path of
    order 4. Any cograph may be constructed, starting from the single vertex
    graph, by a sequence of :meth:`sage.graphs.graph.Graph.join` and
    :meth:`sage.graphs.generic_graph.GenericGraph.disjoint_union` operations.
    See the :wikipedia:`Cograph` for more details.

    This method implements the generator of all cographs of order `n` proposed
    in [JPD2018]_. The algorithm generates one by one every cotree with `n`
    nodes, and each cotree is generated by using its previous cotree. The time
    to construct the first cotree is `O(n)` and the time spent between two
    consecutive outputs is `O(n)`. Hence, the overall complexity of the
    algorithm is `O(n*M_n)`, where `n` is the number of nodes and `M_n` is the
    total number of cographs with `n` nodes (see :oeis:`A000084`).

    INPUT:

    - ``n`` -- integer larger or equal to 1

    - ``as_graph`` -- boolean (default: ``True``); whether to return graphs or
      the tree data structure encoding the graph

    - ``immutable`` -- boolean (default: ``False``); whether to return an
      immutable or a mutable graph. This parameter is used only when ``as_graph
      is True``.

    EXAMPLES:

    The output can be either cotrees or graphs::

        sage: for t in graphs.cographs(3, as_graph=False):
        ....:     print(t)
        [ J ( 0 ) ( 1 ) ( 2 ) ]
        [ J [ U ( 0 ) ( 1 ) ( 2 ) ] ]
        [ J ( 0 ) [ U ( 1 ) ( 2 ) ] ]
        [ J [ U ( 0 ) [ J ( 1 ) ( 2 ) ] ] ]
        sage: for g in graphs.cographs(3, as_graph=True):
        ....:     print(g.edges(labels=False, sort=True))
        [(0, 1), (0, 2), (1, 2)]
        []
        [(0, 1), (0, 2)]
        [(1, 2)]

    Check that we agree with :oeis:`A000084`::

        sage: [sum(1 for _ in graphs.cographs(n, as_graph=False))
        ....:  for n in range(1, 8)]
        [1, 2, 4, 10, 24, 66, 180]

    TESTS::

        sage: g = next(graphs.cographs(2, as_graph=True, immutable=False))
        sage: g.is_immutable()
        False
        sage: g = next(graphs.cographs(2, as_graph=True, immutable=True))
        sage: g.is_immutable()
        True
        sage: next(graphs.cographs(0))
        Traceback (most recent call last):
        ...
        ValueError: parameter n must be at least >= 1
    """
    if n < 1:
        raise ValueError('parameter n must be at least >= 1')
    if as_graph:
        def func(T):
            return tree_to_graph(T, immutable=immutable)
    else:
        def func(T):
            return T

    if n == 1:
        T = CoTree('J')
        B = CoTree('U')
        B.add_child(CoTree(0))
        T.add_child(B)
        yield func(T)
        return

    # Construct the minimum tree
    T = CoTree(n)
    for j in range(n):
        T.add_child(CoTree(1))

    while True:
        # T corresponds to 2 cotrees: one with 'U' root and one with 'J' root
        tree_J = CoTree(T.name)
        T.copy_tree(tree_J)
        tree_U = CoTree(T.name)
        T.copy_tree(tree_U)
        # tree_J has root 'J'
        change_label(tree_J, True, [0])
        # tree 0 has root 'U'
        change_label(tree_U, False, [0])
        tree_UU = CoTree('J')
        tree_UU.add_child(tree_U)
        yield func(tree_J)
        yield func(tree_UU)
        if not next_tree(T):
            break


def change_label(tree, status, counter):
    r"""
    Set the names of the nodes of ``tree``.

    This is a helper method to method :meth:`cographs`.

    The input ``tree`` is such that each node has as label its number of
    children.  This method changes the label of each node so that a parallel
    node is labeled 'U', a series node is labeled 'J' and a leaf node gets a
    unique number.

    INPUT:

    - ``tree`` -- the tree to relabel

    - ``status`` -- boolean; used to to detect series (``True``) and parallel
      (``False``) internal nodes

    - ``counter`` -- list; the first integer of the list is used to assign a
      unique number to the leaves of the tree

    EXAMPLES::

        sage: next(graphs.cographs(4, as_graph=True)).vertices()  # indirect doctest
        [0, 1, 2, 3]
    """
    if tree.name != 1:
        tree.name = 'J' if status else 'U'
    else:
        tree.name = counter[0]
        counter[0] += 1
    for child in tree.children:
        if child is not None:
            change_label(child, not status, counter)


def tree_to_graph(tree, immutable=False):
    r"""
    Return the cograph represented by ``tree``.

    This is a helper method to method :meth:`cographs`.

    EXAMPLES::

        sage: for t in graphs.cographs(3, as_graph=True):  # indirect doctest
        ....:     print(t.edges(labels=False, sort=True))
        [(0, 1), (0, 2), (1, 2)]
        []
        [(0, 1), (0, 2)]
        [(1, 2)]
    """
    from sage.graphs.graph import Graph
    g = Graph()
    _tree_to_graph_rec(tree, g)
    return g.copy(immutable=True) if immutable else g


def _tree_to_graph_rec(tree, g):
    r"""
    Add recursively one by one the vertices and edges of ``tree`` to ``g``.

    This is a helper method to method :meth:`tree_to_graph`.

    EXAMPLES::

        sage: for t in graphs.cographs(3, as_graph=True):  # indirect doctest
        ....:     print(t.edges(labels=False, sort=True))
        [(0, 1), (0, 2), (1, 2)]
        []
        [(0, 1), (0, 2)]
        [(1, 2)]
    """
    for child in tree.children:
        _tree_to_graph_rec(child, g)
    if tree.name not in ['J', 'U']:
        tree.info = 'v'
        g.add_vertex(tree.name)
        _find_neighbors(tree, g)
        tree.reset_info()


def _find_neighbors(tree, g):
    r"""
    Identify the neighbors of node ``tree`` in ``g`` and add needed edges.

    This is a helper method to method :meth:`tree_to_graph`.

    EXAMPLES::

        sage: for t in graphs.cographs(3, as_graph=True):  # indirect doctest
        ....:     print(t.edges(labels=False, sort=True))
        [(0, 1), (0, 2), (1, 2)]
        []
        [(0, 1), (0, 2)]
        [(1, 2)]
    """
    ancestor = tree.parent
    ancestor.info = 'v'
    while ancestor is not None:
        if ancestor.name == 'J':
            for sibling in ancestor.children:
                if sibling != tree and sibling.info != 'v':
                    _add_edge(tree, sibling, g)
        elif ancestor.name == 'U':
            ancestor.info = 'v'
        ancestor = ancestor.parent


def _add_edge(u, v, g):
    r"""
    Add an edge in `g` between the nodes associated to the cotrees `u` and `v`.

    This is a helper method to method :meth:`tree_to_graph`.

    If `v` is not a leaf, then the method add edges between `u` and all leaves
    of the subtree rooted at `v`.

    The method assumes that `u` is a leaf (not tested).

    EXAMPLES::

        sage: for t in graphs.cographs(3, as_graph=True):  # indirect doctest
        ....:     print(t.edges(labels=False, sort=True))
        [(0, 1), (0, 2), (1, 2)]
        []
        [(0, 1), (0, 2)]
        [(1, 2)]
    """
    if v.name in ['J', 'U']:
        for child in v.children:
            _add_edge(u, child, g)
    else:
        g.add_edge(u.name, v.name)
